import sys

class LiftNode:
	def __init__(self, score, label, name):
		self.score = score
		self.label = label
		self.name = name

def readStratumSet(filename):
	f = open(filename, 'r')
	lines = f.readlines()
	f.close()
	retSet = set()
	for line in lines:
		line = line.strip()
		if not line:
			continue
		# get rid of the period at the end
		line = line.replace('.','')
		retSet.add(line)
	return retSet

# Computes the lift AUC of a given set of probabilities and labels,
# assuming that they are all for the same stratum.
# @param probs the probabilities given to the examples
# @param labels the class labels, 1 for positive 0 for negative
# @return
def computeLiftAUC(probs, labels):
	nodeList = list()
	for i in range(len(probs)):
		nodeList.append(LiftNode(probs[i], labels[i]))
	return nodeList

# Computes the lift AUC of a given list Sorts the lists in descending order
# Uses trapezoidal rule to compute AUC Returns the area scaled to a
# fractional X-axis [0,1].
# 
# @param list
#            The List<LiftNode> data
# @return double lift AUC score
def computeLiftAUC(nodeList):
	# sort data by their descending score
	nodeList = sorted(nodeList, key=lambda x: x.score, reverse=True)

	# initialize variables
	nbPos = nbInstance = previousNbPos = previousNbInstance = sum = 0;

	# incrementaly compute lift AUC
	for i in range(len(nodeList)):
		nbInstance += 1 # one more instance
		nbPos += nodeList[i].label # assuming POS_CLASS is not changed

		# look ahead to see if end of stretch of same score
		if i < (len(nodeList) - 1) and abs(nodeList[i].score - nodeList[i+1].score) < 0.000001:
			continue

		# not same stretch, compute lift AUC for this point increment
		# use trapezoidal rule
		sum += (nbInstance - previousNbInstance) * (nbPos + previousNbPos)
		previousNbInstance = nbInstance
		previousNbPos = nbPos

	# divide by nbInstance so that x axis is [0,1]
	return sum / (2.0 * nbInstance)

def readNodes(filename):
	f = open(filename, 'r')
	lines = f.readlines()
	f.close()
	nodes = list()
	for line in lines:
		line = line.strip()
		if not line:
			continue
		parts = line.replace('prob(','').replace(').','').split(',')
		label = POS_CLASS if parts[0] == 'pos' else NEG_CLASS
		name = parts[1]
		prob = float(parts[2])
		nodes.append(LiftNode(prob, label, name))
	return nodes

def computeUplift(targetNodes, otherNodes, quantile, numQuantiles):
	if quantile == 0:
		return 0
	tStart = 0
	tEnd = int((float(quantile) / numQuantiles) * len(targetNodes))
	oStart = 0
	oEnd = int((float(quantile) / numQuantiles) * len(otherNodes))
	pt = countPos(targetNodes, tStart, tEnd)
	po = countPos(otherNodes, oStart, oEnd)
	return pt - po

def countPos(nodes, start, end):
	pos = 0
	for i in range(start, end):
		node = nodes[i]
		if node.label == POS_CLASS:
			pos += 1
	return pos

def splitNodesByStratum(nodes, targetSet, otherSet):
	targetList = list()
	otherList = list()
	for node in nodes:
		if node.name in targetSet:
			targetList.append(node)
		elif node.name in otherSet:
			otherList.append(node)
		else:
			print(node.name + ' not in either stratum')
	return (targetList, otherList)

def main(args):
	# sort the nodes by descending score
	nodes = sorted(readNodes(args[1]), key=lambda x: x.score, reverse=True)
	targetSet = readStratumSet(args[2])
	otherSet = readStratumSet(args[3])
	targetOtherNodes = splitNodesByStratum(nodes, targetSet, otherSet)
	tNodes = targetOtherNodes[0]
	oNodes = targetOtherNodes[1]
	quantiles = int(args[4])
	# start by printing the 0 point
	for q in range(quantiles + 1):
		uplift = computeUplift(tNodes, oNodes, q, quantiles)
		prop = float(q) / quantiles
		print('{0:.2f} {1:d}'.format(prop, uplift))
	pass

POS_CLASS = 1
NEG_CLASS = 0
main(sys.argv)

#prob(pos,invasive(113432),0.28582670083164813).