package sayu;
import java.util.HashMap;

public abstract class BNILPSearch implements ILPSearch{

    //number of positive/negative examples for each data division
    protected int trainPos;
    protected int trainNeg;
    protected int tunePos;
    protected int tuneNeg;
    protected int testPos;
    protected int testNeg;

    //Information about the data
    //Information about the attributes
    protected Attribute[] attributes;
    //Train set
    protected byte[][] train;
    //Tune Set
    protected byte[][] tune;
    //Test Set
    protected byte[][] test;
    //The number of user provided featuers
    protected int numPriorFeats = 0;
    //class values for each example in train, tune, test set
    protected int[][] cats;
    //number of values each feature can take on
    protected int[] numValuesFeat;
    //Position of class variable
    protected int outputPosition;

    //Statistics about the run
    protected int numRules = 0;    
    protected int addEvaluated = 0;
    protected int removeEvaluated = 0;
    protected int replaceEvaluated = 0;
    protected int numAddAccepted = 0;
    protected int numRemoveAccepted = 0;
    protected int numReplaceAccepted = 0;
    protected long startTime = 0;
    protected int headCount = 0;

    //information about the predicates in the network
    //The type of arguments of the predicate
    protected int[] headType;
    //The predicate name
    protected int[] predNumber;
    //Where predicates are located in the data set
    protected HashMap<Integer, Integer> predNameMap;
    protected HashMap<Integer, Integer> invPredName;


    protected int initSize = 256;
    //Information about the current move under consideration
    protected int moveType = -1;
    protected int modifiedV1 = -1;
    protected int modifiedV2 = -1;
    protected int modPred1 = -1;
    protected int modPred2 = -1;
    
    //Initial Parameter strength
    protected double initialStrength = 5;
   //Prior for class = 0
    protected double cat0Prior;
    //Prior for class = 1
    protected double cat1Prior;
    //the counts taken when evaluating a new feature
    protected double[][] condCnts = new double[2][2];
  
    //Information about the network structure and learned model
    protected int[][] graph;
    protected int[] inDegree;
    protected int[] outDegree;
    protected CPTByte[] cpts;

    //Scoring information
    protected double threshold = 0.5;
    protected double minRecall =  0.5;

    /**
       This method introduces a new attribute into the network.  It
       relearns a BN incorporating this feature, then evaluates the
       new network on the tune set according to the score function the
       user selected.

       @param trainPosVector value of new attribute for the pos train examples
       @param trainNegVector value of new attribute for the neg train examples
       @param tunePosVector value of new attribute for the pos tune examples
       @param tuneNegVector value of new attribute for the neg tune examples
       @param testPosVector value of new attribute for the pos test examples
       @param testNegVector value of new attribute for the neg test examples
       @param scoreFunction metric for scoring the new network
       @param numValues the number of values the new attribute can take
       @return the score of the model incorporating the new feature
     */
    public abstract double addAttribute(byte[] trainPosVector, byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector, byte[] testNegVector, int scoreFunction, int numValues);



    /**
      This method introduces a new attribute into the network.  It
      relearns a BN incorporating this feature, then evaluates the
      new network on the tune set according to the score function the
      user selected.

      @param trainPosVector value of new attribute for the pos train
examples
      @param trainNegVector value of new attribute for the neg train
examples
      @param tunePosVector value of new attribute for the pos tune examples
      @param tuneNegVector value of new attribute for the neg tune examples
      @param testPosVector value of new attribute for the pos test examples
      @param testNegVector value of new attribute for the neg test examples
      @param scoreFunction metric for scoring the new network
      @param numValues the number of values the new attribute can take
      @return the score of the model incorporating the new feature
    */
    public abstract double[] addAttribute2(byte[] trainPosVector, 
					   byte[] trainNegVector, 
					   byte[] tunePosVector, 
					   byte[] tuneNegVector, 
					   byte[] testPosVector, byte[] testNegVector, int scoreFunction, int numValues);

    /**
       This method removes an attribute from the network.  It
       relearns a BN without this feature, then evaluates the
       new network on the tune set according to the score function the
       user selected.

       @param removePredNum the predicate number to removed from the model
       @param scoreFunction metric for scoring the new network
       @return the score of the model incorporating the new feature
     */
    public abstract double removeAttribute(int removePredNum, int scoreFunction);

    /**
       This method replaces an attribute already in the model. It
       relearns a BN incorporating this feature in place of a previous
       feature, then evaluates the new network on the tune set
       according to the score function the user selected.

       @param trainPosVector value of new attribute for the pos train examples
       @param trainNegVector value of new attribute for the neg train examples
       @param tunePosVector value of new attribute for the pos tune examples
       @param tuneNegVector value of new attribute for the neg tune examples
       @param testPosVector value of new attribute for the pos test examples
       @param testNegVector value of new attribute for the neg test examples
       @param replacePredNum the predicate number to replaced in the model
       @param scoreFunction metric for scoring the new network
       @param numValues the number of values the new attribute can take
       @return the score of the model incorporating the new feature
     */
    public abstract double replaceAttribute(byte[] trainPosVector, byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector, byte[] testNegVector, int replacePos, int scoreFunction, int numValues);


    /**
       This method permanently adds the most recently introduced
       feature into feature set.  Updates the cpts and network
       topology to reflect the change.
     */
    public abstract void commit();

    /**
       Scores the current model on either train, tune or test
       set. Currently, the following score functions are
       available: CLL, AUC-PR, AUC-ROC and Accuracy.

       @param dataSet whether to score the train, tune or test set
       @param scoreFunction the metric to use for scoring the classifier
       @return score of the model
     */
    public abstract double getScore(int dataSet, int scoreFunction);

    /**
       Returns the probabilities according to the current network structure. 
       @param dataSet whether to return the probabilities for the train, tune or test set.
       @return the probability that each example in the desired dataSet is of the positive class.  
     */
    public abstract double[] getProbs(int dataSet);

    /**
       @return number of non-class attributes in the model
     */
    public int getNumRules(){
	return numRules;
    }

    /**
       Called when constructing a new predicate head. 

       @param type an unique id to present the type and order of
       variables in the predicate head being learned.
     */
    public void newHead(int type){
        headCount++;
        headType[numRules] = type;
        predNumber[numRules] = headCount;
	if (invPredName.containsKey(numRules)){
	    Integer val = invPredName.get(numRules);
	    predNameMap.remove(val);
	    invPredName.remove(numRules);
	}
	invPredName.put(numRules, headCount);
	predNameMap.put(headCount, numRules);
        System.err.println("New Head = " + headCount + " Num Rules = " + numRules);
    }

    /**
       Sets the recall threshold to start integrating from when
       calculating the area under the precision-recall curve.

       @param minRecall is the recall level at which to begin calculating the AUC-PR
    */ 
    public void setAUCPRMinRecall(double minRecall){
	this.minRecall = minRecall;
	System.out.println("Min Recall = " + this.minRecall);
    }//end of setAUCPRMinRecall

    /**
       set the threshold for labeling an example as positive. 

       @param threshold the probability above each an example is classified as positive.
     */
    public void setACCThreshold(double threshold){
	if (threshold < 0){
	    this.threshold = trainPos / ((double)(trainPos+trainNeg));
	}
	else{
	    this.threshold = threshold;
	}
    }//end setACCThreshold(double)

    /**
       Matrix representation of the current network.  Note: graph may
       have more nodes than number of attributes, which is done for
       efficiency purposes.

       @return current netweork structure
    */
    public int[][] getGraph(){
	return graph;
    }//end getGraph()

    /**
       Initializes the class variables and sets up the infrastructure
       for incrementally building a BN based classifier.

       @param trainPos the number of positive examples in the train set.
       @param trainNeg the number of negative examples in the train set.
       @param tunePos the number of positive examples in the tune set.
       @param tuneNeg the number of negative examples in the tune set.
       @param testPos the number of positive examples in the test set.
       @param testNeg the number of negative examples in the test set.
     */
    protected void init(int trainPos,int trainNeg, int tunePos, int tuneNeg, int testPos, int testNeg ){
	startTime = System.currentTimeMillis();
        this.trainPos = trainPos;
        this.trainNeg = trainNeg;
        this.tunePos = tunePos;
        this.tuneNeg = tuneNeg;
        this.testPos = testPos;
        this.testNeg = testNeg;
        
        outputPosition = initSize-1;
        

        cpts = new CPTByte[initSize];

	headType = new int[initSize];
        predNumber = new int[initSize];
	predNameMap = new HashMap<Integer, Integer>();
	invPredName = new HashMap<Integer, Integer>();

        train = new byte[trainPos+trainNeg][initSize];
        tune = new byte[tunePos+tuneNeg][initSize];
        test = new byte[testPos+testNeg][initSize];
        cats = new int[3][];

        cats[0] = new int[train.length];
        cats[1] = new int[tune.length];
        cats[2] = new int[test.length];

        attributes = new Attribute[initSize];
        numValuesFeat = new int[initSize];//numPriorFeats];\
	for(int ix = 0; ix < train.length; ix++){
	    byte cat = (ix >= trainPos) ? (byte)0 : (byte)1;
	    train[ix][outputPosition] = cat;
	    cats[0][ix] = cat;
	}

	for(int ix = 0; ix < tune.length; ix++){
	    byte cat = (ix >= tunePos) ? (byte)0 : (byte)1;
	    tune[ix][outputPosition] = cat;
	    cats[1][ix] = cat;
	}
	for(int ix = 0; ix < test.length; ix++){
	    byte cat = (ix >= testPos) ? (byte)0 : (byte)1;
	    test[ix][outputPosition] = cat;
	    cats[2][ix] = cat;
	}

        attributes[outputPosition] = new VariableConsecInt("Class", 2);
        cpts[outputPosition] = new CPTByte(outputPosition, new int[0], 
					   attributes, train);

	cat0Prior = cpts[outputPosition].getLogProb(train[train.length-1], null, -1);
        cat1Prior = cpts[outputPosition].getLogProb(train[0], null, -1);
	graph = new int[initSize][initSize];

    }

    /**
       Creates the condition distribution of this attribute based on
       the class variable (note all the training data is used to
       compute the distribution).

       @param trainPosVector the bitmap for the positive training examples
       @param trainNegVector the bitmap for the negative training examples
       @param tunePosVector the bitmap for the positive tune examples
       @param tuneNegVector the bitmap for the negative tune examples
       @param testPosVector the bitmap for the positive test examples
       @param testNegVector the bitmap for the negative test examples
     */
    protected void getCondCount(byte[] trainPosVector, byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector, byte[] testNegVector){
        //initialize the cond counts
        int numFeats = numValuesFeat[numRules];
	if (condCnts.length != numFeats){
	    condCnts = new double[numFeats][2];
	}
	double norm = numFeats * 2;
        for(int ix = 0; ix < condCnts.length; ix++){
            for(int jx = 0; jx < condCnts[ix].length; jx++){
                condCnts[ix][jx] = initialStrength / norm;
            }
        }
        /*
	  convert the rule to training set
         */
        for(int ix = 0; ix < trainPos; ix++){
            condCnts[trainPosVector[ix]][1]++;
            train[ix][numRules] = trainPosVector[ix];
        }
        for(int ix = trainPos; ix < train.length ; ix++){
            condCnts[trainNegVector[ix-trainPos]][0]++;
            train[ix][numRules] = trainNegVector[ix-trainPos];
        }
        //add rule to tune set
        for(int ix = 0; ix < tunePos; ix++){
            tune[ix][numRules] = tunePosVector[ix];
        }
        for(int ix = tunePos; ix < tune.length ; ix++){
            tune[ix][numRules] = tuneNegVector[ix-tunePos];
        }
        //add rule to test set
        for(int ix = 0; ix < testPos; ix++){
            test[ix][numRules] = testPosVector[ix];
        }
        for(int ix = testPos; ix < test.length ; ix++){
            test[ix][numRules] = testNegVector[ix-testPos];
        }
    }

    /**
       Doubles the sizes of all of the instance variables. 
    */
    protected void doubleSizes(){
	int newSize = initSize * 2;
	int newOut = newSize-1;

	byte[][] newTrain = new byte[train.length][newSize];
	byte[][] newTune = new byte[tune.length][newSize];
	byte[][] newTest = new byte[test.length][newSize];
	for(int ix = 0; ix < train.length; ix++){
	    System.arraycopy(train[ix], 0, newTrain[ix], 0, numRules);
	    newTrain[ix][newOut] = train[ix][outputPosition];	
	}
	for(int ix = 0; ix < tune.length; ix++){
	    System.arraycopy(tune[ix], 0, newTune[ix], 0, numRules);
	    newTune[ix][newOut] = tune[ix][outputPosition];
	}
	for(int ix = 0; ix < test.length; ix++){
	    System.arraycopy(test[ix], 0, newTest[ix], 0, numRules);
	    newTest[ix][newOut] = test[ix][outputPosition];	
	}
        CPTByte[] newCpts = new CPTByte[newSize];
	System.arraycopy(cpts, 0, newCpts, 0, numRules);
	for(int ix = 0; ix < numRules; ix++){
	    if (cpts[ix] != null){
		cpts[ix].changeParentPosition(outputPosition, newOut);
	    }
	}
	cpts[outputPosition].setThisPos(newOut);
	newCpts[newOut] = cpts[outputPosition];
	int[] newHeadType = new int[newSize];
	System.arraycopy(headType, 0, newHeadType, 0, numRules);
        int[] newPredNumber = new int[newSize];
	System.arraycopy(predNumber, 0, newPredNumber, 0, numRules);
	Attribute[] newAttributes = new Attribute[newSize];
	System.arraycopy(attributes, 0, newAttributes, 0, numRules);
	newAttributes[newOut] = attributes[outputPosition];
        int[] newNumValuesFeat = new int[newSize];
	System.arraycopy(numValuesFeat, 0, newNumValuesFeat, 0, numRules);
	newNumValuesFeat[newOut] = numValuesFeat[outputPosition];
	int[][] newGraph = new int[newSize][newSize];
	for(int ix = 0; ix < graph.length; ix++){
	    System.arraycopy(graph[ix], 0, newGraph[ix], 0, graph[ix].length);
	}

	train = newTrain;
	tune = newTune;
	test = newTest;
	cpts = newCpts;
	attributes = newAttributes;
	numValuesFeat = newNumValuesFeat;
	headType = newHeadType;
	predNumber = newPredNumber;
	graph = newGraph;
	initSize = newSize;
	outputPosition = newOut;
    }

    /**
       Prints out statistics from the most recent SAYU run
     */
    public void terminate(){
        long curTime = System.currentTimeMillis();
        long tmp = (curTime - startTime);
        long minutes = tmp/ (long)60000;
	double secs = Utils.round(((tmp - 60000 * minutes) / 1000.0), 5);
	System.out.println("Current Run Time: " + minutes + " mins " + 
			   secs + " secs");

	System.out.println("Statistics: ");
	System.out.println("Number of rules in the network: " + numRules);
	System.out.print("Number of rules added by SAYU: ");
	System.out.println((numRules-numPriorFeats));
	System.out.println("Number of rules evaluated: " + (addEvaluated + removeEvaluated + replaceEvaluated));
	System.out.println("Number of accepted moves: " + (numAddAccepted + numRemoveAccepted + numReplaceAccepted));
	System.out.println("Number of add evaluated: " + addEvaluated);
	System.out.println("Number of remove evaluated: " + removeEvaluated);
	System.out.println("Number of replace evaluated: " + replaceEvaluated);

	System.out.println("Number of adds accepted: " + numAddAccepted);
	System.out.println("Number of removes accepted: " + numRemoveAccepted);
	System.out.println("Number of replaces accepted: " + numReplaceAccepted);
    }
}