package sayu;

import java.util.Arrays;

public class DifferentialIncBN extends BNILPSearch {
	
	private BNILPSearch targetStratumModel;
	private BNILPSearch otherStratumModel;
	
	private int modelType;
	
	private int trainPos;
	private int trainNeg;
	private int trainPosStratSepIndex;
	private int trainNegStratSepIndex;
	
	private int tunePos;
	private int tuneNeg;
	private int tunePosStratSepIndex;
	private int tuneNegStratSepIndex;
	
	private int testPos;
	private int testNeg;
	private int testPosStratSepIndex;
	private int testNegStratSepIndex;
	
	//buffers for splitting byte arrays by stratum
	private byte[] trainPosTargetBuf;
	private byte[] trainNegTargetBuf;
	private byte[] tunePosTargetBuf;
	private byte[] tuneNegTargetBuf;
	private byte[] testPosTargetBuf;
	private byte[] testNegTargetBuf;
	private byte[] trainPosOtherBuf;
	private byte[] trainNegOtherBuf;
	private byte[] tunePosOtherBuf;
	private byte[] tuneNegOtherBuf;
	private byte[] testPosOtherBuf;
	private byte[] testNegOtherBuf;
	
	//score buffer for addAttribute2
	private double[] out = new double[2];
	//probability buffers for getProbs
	private double[] trainProbBuf;
	private double[] tuneProbBuf;
	private double[] testProbBuf;
	
	/**
	 * 
	 * @param trainPos
	 *            the number of positive examples in the train set.
	 * @param trainNeg
	 *            the number of negative examples in the train set.
	 * @param tunePos
	 *            the number of positive examples in the tune set.
	 * @param tuneNeg
	 *            the number of negative examples in the tune set.
	 * @param testPos
	 *            the number of positive examples in the test set.
	 * @param testNeg
	 *            the number of negative examples in the test set.
	 * @param trainPosStratSep
	 *            end index of target stratum in positive training
	 * @param tunePosStratSep
	 *            end index of target stratum in positive tuning
	 * @param testPosStratSep
	 *            end index of target stratum in positive testing
	 * @param trainNegStratSep
	 *            end index of target stratum in negative training
	 * @param tuneNegStratSep
	 *            end index of target stratum in negative tuning
	 * @param testNegStratSep
	 *            end index of target stratum in negative testing
	 * @param modelType
	 *            flag indicating TAN or BN
	 */
	public DifferentialIncBN(int trainPos, int trainNeg, int tunePos,
			int tuneNeg, int testPos, int testNeg, int trainPosStratSep,
			int tunePosStratSep, int testPosStratSep, int trainNegStratSep,
			int tuneNegStratSep, int testNegStratSep, int modelType) {
		//System.out.println("**DEBUG** " + trainPos + " " + trainNeg + " " + tunePos + " " + tuneNeg + " " + testPos + " " + testNeg);
		//System.out.println("**DEBUG** " + trainPosStratSep + " " + trainNegStratSep);
		//System.out.println("**DEBUG** " + tunePosStratSep + " " + tuneNegStratSep);
		//System.out.println("**DEBUG** " + testPosStratSep + " " + testNegStratSep);
		//set up the initial values
		this.trainPos = trainPos;
		this.trainNeg = trainNeg;
		this.tunePos = tunePos;
		this.tuneNeg = tuneNeg;
		this.testPos = testPos;
		this.testNeg = testNeg;
		this.trainPosStratSepIndex = trainPosStratSep;
		this.tunePosStratSepIndex = tunePosStratSep;
		this.testPosStratSepIndex = testPosStratSep;
		this.trainNegStratSepIndex = trainNegStratSep;
		this.tuneNegStratSepIndex = tuneNegStratSep;
		this.testNegStratSepIndex = testNegStratSep;
		//are we doing TAN or NB?
		this.modelType = modelType;
		//initialize the models for the different strata
		//target stratum comes first in order
		this.targetStratumModel = this.modelType == Constants.TAN ?
				new IncTan(trainPosStratSep, trainNegStratSep,
						tunePosStratSep, tuneNegStratSep, testPosStratSep,
						testNegStratSep) :
				new IncNB(trainPosStratSep, trainNegStratSep,
						tunePosStratSep, tuneNegStratSep, testPosStratSep,
						testNegStratSep);
		//other stratum is second
		this.otherStratumModel = this.modelType == Constants.TAN ?
				new IncTan(trainPos - trainPosStratSep,
						trainNeg - trainNegStratSep, tunePos - tunePosStratSep,
						tuneNeg - tuneNegStratSep, testPos - testPosStratSep,
						testNeg - testNegStratSep) :
				new IncNB(trainPos - trainPosStratSep,
						trainNeg - trainNegStratSep, tunePos - tunePosStratSep,
						tuneNeg - tuneNegStratSep, testPos - testPosStratSep,
						testNeg - testNegStratSep);
		//set up the buffers
		//target stratum byte buffers
		this.trainPosTargetBuf = new byte[trainPosStratSep];
		this.trainNegTargetBuf = new byte[trainNegStratSep];
		this.tunePosTargetBuf = new byte[tunePosStratSep];
		this.tuneNegTargetBuf = new byte[tuneNegStratSep];
		this.testPosTargetBuf = new byte[testPosStratSep];
		this.testNegTargetBuf = new byte[testNegStratSep];
		//other stratum byte buffers
		this.trainPosOtherBuf = new byte[trainPos - trainPosStratSep];
		this.trainNegOtherBuf = new byte[trainNeg - trainNegStratSep];
		this.tunePosOtherBuf = new byte[tunePos - tunePosStratSep];
		this.tuneNegOtherBuf = new byte[tuneNeg - tuneNegStratSep];
		this.testPosOtherBuf = new byte[testPos - testPosStratSep];
		this.testNegOtherBuf = new byte[testNeg - testNegStratSep];
		//probability buffers
		this.trainProbBuf = new double[trainPos + trainNeg];
		this.tuneProbBuf = new double[tunePos + tuneNeg];
		this.testProbBuf = new double[testPos + testNeg];
	}

	@Override
	public double addAttribute(byte[] trainPosVector, byte[] trainNegVector,
			byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector,
			byte[] testNegVector, int scoreFunction, int numValues) {
		//System.out.println("*DEBUG* byte vectors are equal: " + Arrays.equals(trainPosVector, tunePosVector));
		//first fill our byte buffers for each of the strata
		this.fillStratByteBuffers(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector);
		double score = 
				this.targetStratumModel.addAttribute(
					this.trainPosTargetBuf,	this.trainNegTargetBuf,
					this.tunePosTargetBuf, this.tuneNegTargetBuf,
					this.testPosTargetBuf, this.testNegTargetBuf,
					scoreFunction, numValues) -
				this.otherStratumModel.addAttribute(
					this.trainPosOtherBuf,	this.trainNegOtherBuf,
					this.tunePosOtherBuf, this.tuneNegOtherBuf,
					this.testPosOtherBuf, this.testNegOtherBuf,
					scoreFunction, numValues);
		return score;
	}

	@Override
	public double[] addAttribute2(byte[] trainPosVector, byte[] trainNegVector,
			byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector,
			byte[] testNegVector, int scoreFunction, int numValues) {
		//System.out.println("*DEBUG* byte vectors are equal: " + Arrays.equals(trainPosVector, tunePosVector));
		//first fill our byte buffers for each of the strata
		this.fillStratByteBuffers(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector);
		double[] targetOut =
				this.targetStratumModel.addAttribute2(
					this.trainPosTargetBuf,	this.trainNegTargetBuf,
					this.tunePosTargetBuf, this.tuneNegTargetBuf,
					this.testPosTargetBuf, this.testNegTargetBuf,
					scoreFunction, numValues);
		double[] otherOut =
				this.otherStratumModel.addAttribute2(
					this.trainPosOtherBuf,	this.trainNegOtherBuf,
					this.tunePosOtherBuf, this.tuneNegOtherBuf,
					this.testPosOtherBuf, this.testNegOtherBuf,
					scoreFunction, numValues);
		this.out[0] = targetOut[0] - otherOut[0];
		this.out[1] = targetOut[1] - otherOut[1];
		return this.out;
	}

	@Override
	public double removeAttribute(int removePredNum, int scoreFunction) {
		//should just pass along to the models
		double score =
		this.targetStratumModel.removeAttribute(removePredNum, scoreFunction) - 
		this.otherStratumModel.removeAttribute(removePredNum, scoreFunction);
		return score;
	}

	@Override
	public double replaceAttribute(byte[] trainPosVector,
			byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector,
			byte[] testPosVector, byte[] testNegVector, int replacePredNum,
			int scoreFunction, int numValues) {
		//first fill our byte buffers for each of the strata
		this.fillStratByteBuffers(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector);
		double score = 
				this.targetStratumModel.replaceAttribute(
					this.trainPosTargetBuf,	this.trainNegTargetBuf,
					this.tunePosTargetBuf, this.tuneNegTargetBuf,
					this.testPosTargetBuf, this.testNegTargetBuf,
					replacePredNum, scoreFunction, numValues) -
				this.otherStratumModel.replaceAttribute(
					this.trainPosOtherBuf,	this.trainNegOtherBuf,
					this.tunePosOtherBuf, this.tuneNegOtherBuf,
					this.testPosOtherBuf, this.testNegOtherBuf,
					replacePredNum, scoreFunction, numValues);
		return score;
	}

	@Override
	public void commit() {
		//just tell both models to commit
		this.targetStratumModel.commit();
		this.otherStratumModel.commit();
	}

	@Override
	public double getScore(int dataSet, int scoreFunction) {
		//right now, we just want to get the difference in score between models
		//the intent is for scoreFunction to always be lift AUC, but that's not
		//strictly necessary
		return this.targetStratumModel.getScore(dataSet, scoreFunction) -
				this.otherStratumModel.getScore(dataSet, scoreFunction);
	}

	@Override
	public double[] getProbs(int dataSet) {
		//first select the correct probability buffer
		double[] probBuf = null;
		switch (dataSet) {
			case Constants.TRAIN_SET: probBuf = this.trainProbBuf; break;
			case Constants.TUNE_SET: probBuf = this.tuneProbBuf; break;
			case Constants.TEST_SET: probBuf = this.testProbBuf; break;
			default: System.err.println("getProbs called with invalid dataset");
		}
		//now get the actual probabilities in the buffer
		double[] targetProbs = this.targetStratumModel.getProbs(dataSet);
		double[] otherProbs = this.otherStratumModel.getProbs(dataSet);
		System.arraycopy(targetProbs, 0, probBuf, 0, targetProbs.length);
		System.arraycopy(otherProbs, 0, probBuf, targetProbs.length,
				otherProbs.length);
		return probBuf;
	}
	
	private void fillStratByteBuffers(byte[] trainPosVector,
			byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector,
			byte[] testPosVector, byte[] testNegVector) {
		//training
		//target strat is first in order
		System.arraycopy(trainPosVector, 0, this.trainPosTargetBuf, 0,
				this.trainPosStratSepIndex);
		System.arraycopy(trainNegVector, 0, this.trainNegTargetBuf, 0,
				this.trainNegStratSepIndex);
		//other strat is second
		System.arraycopy(trainPosVector, this.trainPosStratSepIndex,
				this.trainPosOtherBuf, 0,
				this.trainPos - this.trainPosStratSepIndex);
		System.arraycopy(trainNegVector, this.trainNegStratSepIndex,
				this.trainNegOtherBuf, 0,
				this.trainNeg - this.trainNegStratSepIndex);
		
		//tuning
		//target strat is first in order
		System.arraycopy(tunePosVector, 0, this.tunePosTargetBuf, 0,
				this.tunePosStratSepIndex);
		System.arraycopy(tuneNegVector, 0, this.tuneNegTargetBuf, 0,
				this.tuneNegStratSepIndex);
		//other strat is second
		System.arraycopy(tunePosVector, this.tunePosStratSepIndex,
				this.tunePosOtherBuf, 0,
				this.tunePos - this.tunePosStratSepIndex);
		System.arraycopy(tuneNegVector, this.tuneNegStratSepIndex,
				this.tuneNegOtherBuf, 0,
				this.tuneNeg - this.tuneNegStratSepIndex);
		
		//testing
		//target strat is first in order
		System.arraycopy(testPosVector, 0, this.testPosTargetBuf, 0,
				this.testPosStratSepIndex);
		System.arraycopy(testNegVector, 0, this.testNegTargetBuf, 0,
				this.testNegStratSepIndex);
		//other strat is second
		System.arraycopy(testPosVector, this.testPosStratSepIndex,
				this.testPosOtherBuf, 0,
				this.testPos - this.testPosStratSepIndex);
		System.arraycopy(testNegVector, this.testNegStratSepIndex,
				this.testNegOtherBuf, 0,
				this.testNeg - this.testNegStratSepIndex);
	}

}
