package sayu;

import java.util.*;

/**
 * @author Jesse Davis
 * 
 *         This class is for incrementally building a Tree Augmented Naive Bayes
 *         model. It stores the constructed train, tune and test sets. It also
 *         stores the CPT for each node in the network.
 * 
 *         (c) 2010
 */
public class IncTan extends BNILPSearch {

	// Edges between all pairs of attributes in the model
	private Edge[][] allEdges;
	// The edges in the current model
	private Edge[] bestEdges;

	// Edges in model after adding a new attribute
	private Edge[] currentEdges;
	// Edges constructed between new attribute and prior attributes
	private Edge[] evalEdges;
	// Edges in model after remove an attribute
	private Edge[] removeEdges;
	private Edge[] replaceEdges;

	// for storing the prob(val | class) for each feature
	private double[][][] condProbs;

	// Arrays that are reused multiple times, so just create once
	// for computing the conditional distribution
	private double[] totals = new double[2];
	private int[] classPar = new int[1];
	private int[] twoPar = new int[2];

	// The network under consideration
	private int[][] evalGraph;
	private int[] evalInDegree;
	private int[] evalOutDegree;
	// CPTs for the network under consideration
	private CPTByte[] evalCPTs;
	// The non class parent of each attribute in the network being considered
	private int[] evalParents;

	// The non class parent of each attribute in the current model
	private int[] parents;

	private double[] out = new double[2];

	/**
	 * Constructs a TAN model with only the class variable.
	 * 
	 * @param trainPos
	 *            the number of positive examples in the train set.
	 * @param trainNeg
	 *            the number of negative examples in the train set.
	 * @param tunePos
	 *            the number of positive examples in the tune set.
	 * @param tuneNeg
	 *            the number of negative examples in the tune set.
	 * @param testPos
	 *            the number of positive examples in the test set.
	 * @param testNeg
	 *            the number of negative examples in the test set.
	 */
	public IncTan(int trainPos, int trainNeg, int tunePos, int tuneNeg,
			int testPos, int testNeg) {
		init(trainPos, trainNeg, tunePos, tuneNeg, testPos, testNeg);
	}// end IncTan(int, int, int, int, int, int)

	/**
	 * Constructs a TAN model based on an initial feature set
	 * 
	 * @param trainPos
	 *            the number of positive examples in the train set.
	 * @param trainNeg
	 *            the number of negative examples in the train set.
	 * @param tunePos
	 *            the number of positive examples in the tune set.
	 * @param tuneNeg
	 *            the number of negative examples in the tune set.
	 * @param testPos
	 *            the number of positive examples in the test set.
	 * @param testNeg
	 *            the number of negative examples in the test set.
	 * @param initTrain
	 *            a initial set of train examples
	 * @param initTune
	 *            a initial set of tune examples
	 * @param initTest
	 *            a initial set of test examples
	 * @param initAttribute
	 *            a initial feature set
	 * @param initOutputPos
	 *            is the index of the output position in the initial feature set
	 */
	public IncTan(int trainPos, int trainNeg, int tunePos, int tuneNeg,
			int testPos, int testNeg, byte[][] initTrain, byte[][] initTune,
			byte[][] initTest, Attribute[] initAttributes, int initOutputPos) {
		// make sure that we can hold all the data
		while (initSize < (initTrain[0].length - 1)) {
			initSize = initSize * 2;
		}
		// initialize the model
		init(trainPos, trainNeg, tunePos, tuneNeg, testPos, testNeg);
		// seed the datasets with the initial features
		Utils.copyByteArray(initTrain, train, initOutputPos);
		Utils.copyByteArray(initTune, tune, initOutputPos);
		Utils.copyByteArray(initTest, test, initOutputPos);
		// initialize attributes, numValuesFeat21
		int idx = 0;
		for (int ix = 0; ix < initAttributes.length; ix++) {
			if (ix != initOutputPos) {
				attributes[idx] = initAttributes[ix];
				numValuesFeat[idx] = attributes[idx].getNumValues();
				idx++;
			}
		}
		numRules = initTrain[0].length;

		// Construct Edges
		for (int ix = 0; ix < numRules; ix++) {
			condProbs[ix] = Utils.condProb(train, ix, outputPosition,
					numValuesFeat[ix], 2, initialStrength, false);
		}
		Edge[] tree = new Edge[((numRules - 1) * (numRules - 2)) / 2];
		idx = 0;
		for (int ix = 0; ix < (numRules - 1); ix++) {
			for (int jx = (ix + 1); jx < (numRules - 1); jx++) {
				// ignore the position to be removed
				allEdges[ix][jx] = getIP(ix, jx, condProbs[ix], condProbs[jx]);
				allEdges[jx][ix] = allEdges[ix][jx];
				tree[idx] = allEdges[ix][jx];
				idx++;
			}
		}

		// Build the TAN network
		initEvalGraph(numRules - 1);
		bestEdges = new Edge[numRules - 2];
		GraphUtils.kruskals(tree, graph, outDegree, bestEdges);
		GraphUtils.makeDirected(graph, 0, outDegree, -1, inDegree, parents);
		// learn cpts
		for (int ix = 0; ix < numRules; ix++) {
			if (inDegree[ix] == 0) {
				cpts[ix] = new CPTByte(ix, classPar, attributes, train);
			} else {
				twoPar[1] = outputPosition;
				twoPar[0] = parents[ix];
				cpts[ix] = new CPTByte(ix, twoPar, attributes,
						allEdges[ix][parents[ix]]);
			}
		}
	}// end

	/**
	 * Initializes the class variables and sets up the infrastructure for
	 * incrementally building a TAN network.
	 * 
	 * @param trainPos
	 *            the number of positive examples in the train set.
	 * @param trainNeg
	 *            the number of negative examples in the train set.
	 * @param tunePos
	 *            the number of positive examples in the tune set.
	 * @param tuneNeg
	 *            the number of negative examples in the tune set.
	 * @param testPos
	 *            the number of positive examples in the test set.
	 * @param testNeg
	 *            the number of negative examples in the test set.
	 */
	protected void init(int trainPos, int trainNeg, int tunePos, int tuneNeg,
			int testPos, int testNeg) {

		super.init(trainPos, trainNeg, tunePos, tuneNeg, testPos, testNeg);

		// for building CPTs
		classPar[0] = outputPosition;
		twoPar[1] = outputPosition;
		// set up info about parents
		condProbs = new double[initSize][][];
		parents = new int[initSize];

		// Speculatively network structure and info
		evalCPTs = new CPTByte[initSize];
		evalParents = new int[initSize];
		evalGraph = new int[initSize][initSize];
		evalInDegree = new int[initSize];
		evalOutDegree = new int[initSize];

		for (int ix = 0; ix < parents.length; ix++) {
			parents[ix] = Constants.NO_PARENT;
			evalParents[ix] = Constants.NO_PARENT;
		}

		//
		allEdges = new Edge[initSize][initSize];

		// init graph
		inDegree = new int[initSize];
		outDegree = new int[initSize];
		for (int ix = 0; ix < graph.length - 1; ix++) {
			graph[ix][outputPosition] = Constants.IN_EDGE;
			evalGraph[ix][outputPosition] = Constants.IN_EDGE;
			graph[outputPosition][ix] = Constants.OUT_EDGE;
			evalGraph[outputPosition][ix] = Constants.OUT_EDGE;
		}
	}// end init(int, int, int, int, int, int)

	/**
	 * This method introduces a new attribute into the network. It relearns a
	 * TAN network incorporating this feature, then evaluates the new network on
	 * the tune set according to the score function the user selected.
	 * 
	 * @param trainPosVector
	 *            value of new attribute for the pos train examples
	 * @param trainNegVector
	 *            value of new attribute for the neg train examples
	 * @param tunePosVector
	 *            value of new attribute for the pos tune examples
	 * @param tuneNegVector
	 *            value of new attribute for the neg tune examples
	 * @param testPosVector
	 *            value of new attribute for the pos test examples
	 * @param testNegVector
	 *            value of new attribute for the neg test examples
	 * @param scoreFunction
	 *            metric for scoring the new network
	 * @param numValues
	 *            the number of values the new attribute can take
	 * @return the score of the model incorporating the new feature
	 */
	public double addAttribute(byte[] trainPosVector, byte[] trainNegVector,
			byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector,
			byte[] testNegVector, int scoreFunction, int numValues) {
		moveType = Constants.ADD;
		addEvaluated++;
		numValuesFeat[numRules] = numValues;
		// get the conditional counts
		getCondCount(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector);

		attributes[numRules] = new VariableConsecInt("R" + numRules, numValues);
		if (numRules == 0) {
			evalCPTs[0] = new CPTByte(0, classPar, attributes, train);
		} else if (numRules == 1) {
			twoPar[0] = 0;
			twoPar[1] = outputPosition;
			evalParents[0] = Constants.NO_PARENT;
			evalParents[1] = 0;
			currentEdges[0] = getIP(0, 1, condProbs[0], condProbs[1]);
			allEdges[0][1] = currentEdges[0];
			allEdges[1][0] = currentEdges[0];
			evalCPTs[0] = cpts[0];
			evalCPTs[1] = new CPTByte(1, twoPar, attributes, currentEdges[0]);
			evalGraph[0][1] = Constants.OUT_EDGE;
			evalGraph[1][0] = Constants.IN_EDGE;
		} else {
			// construct edge between new attribute and each prev attribute
			evalEdges = new Edge[numRules];
			for (int ix = 0; ix < numRules; ix++) {
				evalEdges[ix] = getIP(ix, numRules, condProbs[ix],
						condProbs[numRules]);
				allEdges[ix][numRules] = evalEdges[ix];
				allEdges[numRules][ix] = evalEdges[ix];
			}
			// learn the new network
			initEvalGraph(numRules + 1);
			GraphUtils.kruskals(bestEdges, evalEdges, currentEdges, evalGraph,
					evalOutDegree);
			GraphUtils.makeDirected(evalGraph, 0, evalOutDegree, -1,
					evalInDegree, evalParents);

			// add the class variable as a parent
			for (int ix = 0; ix < evalGraph.length - 1; ix++) {
				evalGraph[ix][evalGraph.length - 1] = Constants.IN_EDGE;
				evalGraph[evalGraph.length - 1][ix] = Constants.OUT_EDGE;
			}
			// learn the cpts for each node
			for (int ix = 0; ix <= numRules; ix++) {
				// Re-use old CPT
				if (evalInDegree[ix] == 0 && ix < numRules
						&& cpts[ix].getNumParents() == 1) {
					evalCPTs[ix] = cpts[ix];
				} else if (evalInDegree[ix] == 0) {
					// May be possible to optimize
					evalCPTs[ix] = new CPTByte(ix, classPar, attributes, train);
				}
				// Re-use old CPT
				else if (parents[ix] == evalParents[ix] && ix < numRules
						&& cpts[ix].getNumParents() == 2) {
					evalCPTs[ix] = cpts[ix];
				}
				// Learn new CPT
				else {
					twoPar[1] = outputPosition;
					twoPar[0] = evalParents[ix];
					evalCPTs[ix] = new CPTByte(ix, twoPar, attributes,
							allEdges[ix][evalParents[ix]]);
				}
			}
		}
		return getScore(Constants.TUNE_SET, scoreFunction, evalCPTs,
				numRules + 1, -1, -1);
	}// end addAttribute(byte[], byte[], byte[], byte[], byte[], byte[], int,
		// int)

	/**
	 * This method introduces a new attribute into the network. It relearns a BN
	 * incorporating this feature, then evaluates the new network on the tune
	 * set according to the score function the user selected.
	 * 
	 * @param trainPosVector
	 *            value of new attribute for the pos train examples
	 * @param trainNegVector
	 *            value of new attribute for the neg train examples
	 * @param tunePosVector
	 *            value of new attribute for the pos tune examples
	 * @param tuneNegVector
	 *            value of new attribute for the neg tune examples
	 * @param testPosVector
	 *            value of new attribute for the pos test examples
	 * @param testNegVector
	 *            value of new attribute for the neg test examples
	 * @param scoreFunction
	 *            metric for scoring the new network
	 * @param numValues
	 *            the number of values the new attribute can take
	 * @return the score of the model incorporating the new feature
	 */
	public double[] addAttribute2(byte[] trainPosVector, byte[] trainNegVector,
			byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector,
			byte[] testNegVector, int scoreFunction, int numValues) {
		out[1] = addAttribute(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector, scoreFunction,
				numValues);
		out[0] = getScore(Constants.TRAIN_SET, scoreFunction, evalCPTs,
				numRules + 1, -1, -1);
		return out;
	}

	/**
	 * This method removes an attribute from the network. It relearns a TAN
	 * network without this feature, then evaluates the new network on the tune
	 * set according to the score function the user selected.
	 * 
	 * @param removePredNum
	 *            the predicate number to removed from the model
	 * @param scoreFunction
	 *            metric for scoring the new network
	 * @return the score of the model incorporating the new feature
	 */
	public double removeAttribute(int removePredNum, int scoreFunction) {
		modifiedV1 = predNameMap.get(removePredNum);
		int removePos = modifiedV1;
		modPred1 = removePredNum;
		removeEvaluated++;
		moveType = Constants.REMOVE;
		/*
		 * Fast remove not always available, so look at edges between all
		 * possible pairs of nodes in network
		 */
		Edge[] tree = new Edge[((numRules - 1) * (numRules - 2)) / 2];
		int pos = 0;
		for (int ix = 0; ix < numRules; ix++) {
			// ignore the position to be removed
			if (ix != removePos) {
				for (int jx = (ix + 1); jx < numRules; jx++) {
					// ignore the position to be removed
					if (jx != removePos) {
						tree[pos] = allEdges[ix][jx];
						pos++;
					}
				}// end jx loop
			}// end if ix != removePos
		}// end of ix loop

		// Build the TAN network
		initEvalGraph(numRules + 1);
		GraphUtils.kruskals(tree, evalGraph, evalOutDegree, removeEdges);
		int startPos = (modifiedV1 == 0) ? 1 : 0;
		GraphUtils.makeDirected(evalGraph, startPos, evalOutDegree, -1,
				evalInDegree, evalParents);
		// Build the CPTs
		for (int ix = 0; ix < numRules; ix++) {
			// ignore the position to be removed
			if (ix != removePos) {
				if (evalInDegree[ix] == 0 && ix < numRules
						&& cpts[ix].getNumParents() == 1) {
					// FLAG - is this still true?
					evalCPTs[ix] = cpts[ix];
				} else if (evalInDegree[ix] == 0) {
					// May be possible to optimize
					evalCPTs[ix] = new CPTByte(ix, classPar, attributes, train);
				} else if (parents[ix] == evalParents[ix] && ix < numRules
						&& cpts[ix].getNumParents() == 2) {
					// FLAG - is this still true?
					evalCPTs[ix] = cpts[ix];
				} else {
					twoPar[1] = outputPosition;
					twoPar[0] = evalParents[ix];
					evalCPTs[ix] = new CPTByte(ix, twoPar, attributes,
							allEdges[ix][evalParents[ix]]);
				}
			}// end if ix != removePos
		}// end of ix loop
		return getScore(Constants.TUNE_SET, scoreFunction, evalCPTs, numRules,
				modifiedV1, -1);
	}// end removeAttribute(int, int)

	/**
	 * This method replaces an attribute already in the model. It relearns a TAN
	 * network incorporating this feature in place of a previous feature, then
	 * evaluates the new network on the tune set according to the score function
	 * the user selected.
	 * 
	 * @param trainPosVector
	 *            value of new attribute for the pos train examples
	 * @param trainNegVector
	 *            value of new attribute for the neg train examples
	 * @param tunePosVector
	 *            value of new attribute for the pos tune examples
	 * @param tuneNegVector
	 *            value of new attribute for the neg tune examples
	 * @param testPosVector
	 *            value of new attribute for the pos test examples
	 * @param testNegVector
	 *            value of new attribute for the neg test examples
	 * @param replacePredNum
	 *            the predicate number to replaced in the model
	 * @param scoreFunction
	 *            metric for scoring the new network
	 * @param numValues
	 *            the number of values the new attribute can take
	 * @return the score of the model incorporating the new feature
	 */
	public double replaceAttribute(byte[] trainPosVector,
			byte[] trainNegVector, byte[] tunePosVector, byte[] tuneNegVector,
			byte[] testPosVector, byte[] testNegVector, int replacePredNum,
			int scoreFunction, int numValues) {
		try {
			modifiedV1 = predNameMap.get(replacePredNum);
			int replacePos = modifiedV1;
			modPred1 = replacePredNum;
			attributes[numRules] = new VariableConsecInt("R" + numRules,
					numValues);

			replaceEvaluated++;
			moveType = Constants.REPLACE;
			numValuesFeat[numRules] = numValues;
			// get the conditional counts and treat as if adding new rule
			getCondCount(trainPosVector, trainNegVector, tunePosVector,
					tuneNegVector, testPosVector, testNegVector);

			Edge[] tree = new Edge[(numRules * (numRules - 1)) / 2];
			int pos = 0;
			for (int ix = 0; ix < numRules; ix++) {
				for (int jx = (ix + 1); jx < numRules; jx++) {
					// create new edge if replace pos is used
					// do pseudo overwrite of old rule
					if (ix == replacePos) {
						tree[pos] = getIP(numRules, jx, condProbs[numRules],
								condProbs[jx]);
						allEdges[numRules][jx] = tree[pos];
						allEdges[jx][numRules] = tree[pos];
						pos++;
					}
					// create new edge if replace pos is used
					// do pseudo overwrite of old rule
					else if (jx == replacePos) {
						tree[pos] = getIP(ix, numRules, condProbs[ix],
								condProbs[numRules]);
						allEdges[ix][numRules] = tree[pos];
						allEdges[numRules][ix] = tree[pos];
						pos++;
					} else {
						tree[pos] = allEdges[ix][jx];
						pos++;
					}
				}
			}
			// build the TAN network
			initEvalGraph(numRules + 1);
			replaceEdges = new Edge[numRules - 1];
			int startPos = (modifiedV1 == 0) ? 1 : 0;
			GraphUtils.kruskals(tree, evalGraph, evalOutDegree, replaceEdges);
			GraphUtils.makeDirected(evalGraph, startPos, evalOutDegree, -1,
					evalInDegree, evalParents);
			// Learn the CPTs
			for (int ix = 0; ix <= numRules; ix++) {
				// FLAG
				if (ix != replacePos) {

					if (evalInDegree[ix] == 0 && ix < numRules
							&& cpts[ix].getNumParents() == 1) {
						evalCPTs[ix] = cpts[ix];
					} else if (evalInDegree[ix] == 0) {
						// May be possible to optimize
						evalCPTs[ix] = new CPTByte(ix, classPar, attributes,
								train);
					} else if (parents[ix] == evalParents[ix] && ix < numRules
							&& cpts[ix].getNumParents() == 2) {
						evalCPTs[ix] = cpts[ix];
					} else {
						twoPar[1] = outputPosition;
						twoPar[0] = evalParents[ix];
						evalCPTs[ix] = new CPTByte(ix, twoPar, attributes,
								allEdges[ix][evalParents[ix]]);
					}
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return getScore(Constants.TUNE_SET, scoreFunction, evalCPTs,
				numRules + 1, modifiedV1, -1);
	}// end replaceAttribute(byte[], byte[], byte[], byte[], byte[], byte[],
		// int, int)

	/**
	 * This method permanently adds the most recently introduced feature into
	 * feature set. Updates the cpts and network topology to reflect the change.
	 */
	public void commit() {
		long curTime = System.currentTimeMillis();
		long tmp = (curTime - startTime);
		long minutes = tmp / (long) 60000;
		double secs = Utils.round(((tmp - 60000 * minutes) / 1000.0), 5);

		if (moveType == Constants.ADD) {
			System.out.print("Added rule: ");
			// increment the number of rules in the network
			numRules++;
			bestEdges = currentEdges;
			numAddAccepted++;
		} else if (moveType == Constants.REMOVE) {
			System.out.print("Removed rule: ");
			// decrement the number of rules in the network
			numRules--;
			bestEdges = removeEdges;
			numRemoveAccepted++;
			// Swap out old rule
			invPredName.remove(numRules);
			predNameMap.remove(modPred1);

			// Flag think about rule/predicate indices
			if (modifiedV1 != numRules) {
				invPredName.put(modifiedV1, predNumber[numRules]);
				predNameMap.put(predNumber[numRules], modifiedV1);
			}
		} else if (moveType == Constants.REPLACE) {
			System.out.print("Replaced rule: ");
			numReplaceAccepted++;
			bestEdges = replaceEdges;
		} else if (moveType == Constants.DISJOIN) {
			numRules--;
		}
		System.out.print(minutes + " mins " + secs + " secs ");
		System.out.println("rules in network " + numRules + " added by SAYU "
				+ (numRules - numPriorFeats) + " evaluated "
				+ (addEvaluated + removeEvaluated + replaceEvaluated));
		if (moveType == Constants.REPLACE || moveType == Constants.REMOVE) {
			predNumber[modifiedV1] = predNumber[numRules];
			headType[modifiedV1] = headType[numRules];
			condProbs[modifiedV1] = condProbs[numRules];
			evalParents[modifiedV1] = evalParents[numRules];
			for (int ix = 0; ix < train.length; ix++) {
				train[ix][modifiedV1] = train[ix][numRules];
			}
			for (int ix = 0; ix < tune.length; ix++) {
				tune[ix][modifiedV1] = tune[ix][numRules];
			}
			for (int ix = 0; ix < test.length; ix++) {
				test[ix][modifiedV1] = test[ix][numRules];
			}
			for (int ix = 0; ix < numRules; ix++) {
				// FLAG: what if ix == modifiedV1?
				if (ix != modifiedV1) {
					if (evalParents[ix] == numRules) {
						evalParents[ix] = modifiedV1;
						evalCPTs[ix].changeParentPosition(numRules, modifiedV1);
					}
					// mod graph
					evalGraph[modifiedV1][ix] = evalGraph[numRules][ix];
					evalGraph[ix][modifiedV1] = evalGraph[ix][numRules];
					// Flip edges
					allEdges[ix][numRules].updateVertex(numRules, modifiedV1);
					allEdges[ix][modifiedV1] = allEdges[ix][numRules];
					allEdges[modifiedV1][ix] = allEdges[ix][numRules];
				}
			}
			evalCPTs[modifiedV1] = evalCPTs[numRules];
			evalCPTs[modifiedV1].setThisPos(modifiedV1);
			evalCPTs[numRules] = null;

			evalInDegree[modifiedV1] = evalInDegree[numRules];
			evalOutDegree[modifiedV1] = evalOutDegree[numRules];
		}
		// update the parents
		System.arraycopy(evalParents, 0, parents, 0, numRules);
		// update cpts
		System.arraycopy(evalCPTs, 0, cpts, 0, numRules);
		// update graph structure
		graph = evalGraph;
		inDegree = evalInDegree;
		outDegree = evalOutDegree;
		// reinit current edges,
		currentEdges = new Edge[numRules];
		if (bestEdges != null) {
			System.out.println("***PRINTING EDGES***");
			System.out.flush();
			for (int ex = 0; ex < bestEdges.length; ex++) {
				System.out.println(bestEdges[ex]);
				System.out.flush();
			}
			System.out.println("***DONE PRINTING EDGES***");
			removeEdges = new Edge[bestEdges.length - 1];
		}

		evalGraph = new int[initSize][initSize];
		evalInDegree = new int[initSize];
		evalOutDegree = new int[initSize];

		//
		moveType = -1;
		modifiedV1 = -1;
		modifiedV2 = -1;
		modPred1 = -1;
		modPred2 = -1;
		if (numRules == (initSize - 1)) {
			doubleSizes();
		}
	}// end commit()

	/**
	 * Creates the condition distribution of this attribute based on the class
	 * variable (note all the training data is used to compute the
	 * distribution).
	 * 
	 * @param trainPosVector
	 *            the bitmap for the positive training examples
	 * @param trainNegVector
	 *            the bitmap for the negative training examples
	 * @param tunePosVector
	 *            the bitmap for the positive tune examples
	 * @param tuneNegVector
	 *            the bitmap for the negative tune examples
	 * @param testPosVector
	 *            the bitmap for the positive test examples
	 * @param testNegVector
	 *            the bitmap for the negative test examples
	 */
	protected void getCondCount(byte[] trainPosVector, byte[] trainNegVector,
			byte[] tunePosVector, byte[] tuneNegVector, byte[] testPosVector,
			byte[] testNegVector) {
		// get the counts, copy feature into train, tune and test sets
		super.getCondCount(trainPosVector, trainNegVector, tunePosVector,
				tuneNegVector, testPosVector, testNegVector);
		// compute the probabilities
		condProbs[numRules] = new double[numValuesFeat[numRules]][2];
		for (int ix = 0; ix < condCnts[0].length; ix++) {
			double total = 0;
			// get totals
			for (int jx = 0; jx < condCnts.length; jx++) {
				total += condCnts[jx][ix];
			}
			// calculate the conditional probabilities
			for (int jx = 0; jx < condCnts.length; jx++) {
				condProbs[numRules][jx][ix] = condCnts[jx][ix] / total;
			}
		}// end of ix loop for getting totals/probs
	}// end of method getCond

	/**
	 * Calculates the conditional mutual information between a1 and a2.
	 * 
	 * @param a1
	 *            an rule already in the network
	 * @param a2
	 *            the rule being evaluated
	 * @param probA1
	 *            the probability(attrib1 | class)
	 * @param probA2
	 *            the probability(newRule | class)
	 * @return weighted [by CMI] edge between a1 and a2
	 */
	protected Edge getIP(int a1, int a2, double[][] probA1, double[][] probA2) {
		double[][][] threeCnts = new double[numValuesFeat[a1]][numValuesFeat[a2]][2];
		double norm = numValuesFeat[a1] * numValuesFeat[a2] * 2;
		// initialize all counts
		for (int ix = 0; ix < threeCnts.length; ix++) {
			for (int jx = 0; jx < threeCnts[ix].length; jx++) {
				for (int kx = 0; kx < threeCnts[ix][jx].length; kx++) {
					threeCnts[ix][jx][kx] = initialStrength / norm;
				}
			}
		}
		// get the counts
		for (int ix = 0; ix < train.length; ix++) {
			threeCnts[train[ix][a1]][train[ix][a2]][train[ix][outputPosition]]++;
		}

		for (int ix = 0; ix < totals.length; ix++) {
			totals[ix] = 0;
		}
		double sumTot = 0;
		// get total and conditional totals over joint counts
		for (int ix = 0; ix < threeCnts.length; ix++) {
			for (int jx = 0; jx < threeCnts[ix].length; jx++) {
				for (int kx = 0; kx < threeCnts[ix][jx].length; kx++) {
					sumTot += threeCnts[ix][jx][kx];
					totals[kx] += threeCnts[ix][jx][kx];
				}
			}
		}// end ix loop

		double sum = 0;
		double tmp1 = 0;
		double tmp2 = 0;
		double tmpSum = 0;
		// calculate the mutual information
		for (int ix = 0; ix < threeCnts.length; ix++) {
			for (int jx = 0; jx < threeCnts[ix].length; jx++) {
				for (int kx = 0; kx < threeCnts[ix][jx].length; kx++) {
					// P(ix,jx,kx)
					tmp1 = (double) threeCnts[ix][jx][kx] / (double) sumTot;
					// P(ix,jx |kx)
					tmp2 = (double) threeCnts[ix][jx][kx] / (double) totals[kx];

					// a1[ix][kx] = P(ix | kx)
					// a1[jx][kx] = P(jx | kx)
					tmpSum = tmp1
							* Math.log(tmp2 / (probA1[ix][kx] * probA2[jx][kx]));
					sum += tmpSum;
				}// end of kx loop (all class values
			}// end of jx loop (all values for attribute at pos2
		}// end of ix loop (all values for attribute at pos1
		return new Edge(a1, a2, sum, threeCnts);
	}// end method getIP(int, int, double[][], double[][]

	/**
	 * Scores the current model on either train, tune or test set. Currently,
	 * the following score functions are available: CLL, AUC-PR, AUC-ROC and
	 * Accuracy.
	 * 
	 * @param dataSet
	 *            whether to score the train, tune or test set
	 * @param scoreFunction
	 *            the metric to use for scoring the classifier
	 * @return score of the model
	 */
	public double getScore(int dataSet, int scoreFunction) {
		return getScore(dataSet, scoreFunction, cpts, numRules, -1, -1);
	}// end of method getScore(int, int)

	/**
	 * Score a model on either train, tune or test set. The model could be a
	 * speculative model. Currently, the following score functions are
	 * available: CLL, AUC-PR, AUC-ROC and Accuracy.
	 * 
	 * @param dataSet
	 *            whether to score the train, tune or test set.
	 * @param scoreFunction
	 *            the metric to use for scoring the classifier.
	 * @param probDist
	 *            is the prob distribution for each node
	 * @param bound
	 *            the number of non-class attributes to consider
	 * @return the current score of the model
	 */
	protected double getScore(int dataSet, int scoreFunction,
			ProbDistribution[] probDist, int bound, int ignoreV1, int ignoreV2) {
		double[] probs = null;
		int posCnt = trainPos;
		int negCnt = trainNeg;
		int idx = 0;
		if (dataSet == Constants.TRAIN_SET) {
			probs = getProbs(train, bound, probDist, ignoreV1, ignoreV2);
		} else if (dataSet == Constants.TUNE_SET) {
			probs = getProbs(tune, bound, probDist, ignoreV1, ignoreV2);
			posCnt = tunePos;
			negCnt = tuneNeg;
			idx = 1;
		} else if (dataSet == Constants.TEST_SET) {
			probs = getProbs(test, bound, probDist, ignoreV1, ignoreV2);
			posCnt = testPos;
			negCnt = testNeg;
			idx = 2;
		}
		if (scoreFunction == Constants.ACC) {
			// FLAG
			int[] roc = ScoreFunctions.fastAccuracy(probs, cats[idx], 0.5);
			return ((double) (roc[Constants.TP] + roc[Constants.TN]))
					/ ((double) (roc[0] + roc[1] + roc[2] + roc[3]));
		} else if (scoreFunction == Constants.PR_AREA) {
			double[] scores = ScoreFunctions.calcPRArea(probs, cats[idx],
					posCnt, negCnt);
			return scores[1];
		} else if (scoreFunction == Constants.ROC_AREA) {
			return ScoreFunctions.calcROCArea(probs, cats[idx]);
		} else if (scoreFunction == Constants.CLL) {
			return ScoreFunctions.cll(probs, cats[idx]);
		} else if (scoreFunction == Constants.X_ENT) {
			return -1;// ScoreFunctions.crossEntropyError(probs, cats[idx]);
		} else if (scoreFunction == Constants.MSQE) {
			return -1;// ScoreFunctions.msqe(probs, cats[idx]);
		} else if (scoreFunction == Constants.LIFT_AREA) {
			return Uplift.computeLiftAUC(probs, this.cats[idx]);
		}
		return -1;
	}// end getScore(int, int, ProbDistribution[], int)

	/**
	 * Calculates the probability that each example in either the train, tune or
	 * test set belongs to the positive class. It uses the current model
	 * structure.
	 * 
	 * @param dataSet
	 *            whether to return the probabilities for the train, tune or
	 *            test set.
	 * @return the Prob(Pos Class) of each example in the desired dataSet
	 */
	public double[] getProbs(int dataSet) {
		if (dataSet == Constants.TRAIN_SET) {
			return getProbs(train, numRules, cpts, -1, -1);
		} else if (dataSet == Constants.TUNE_SET) {
			return getProbs(tune, numRules, cpts, -1, -1);
		} else if (dataSet == Constants.TEST_SET) {
			return getProbs(test, numRules, cpts, -1, -1);
		}
		System.out.println("Data set is invalid");
		return null;
	}// end getProbs(int)

	/**
	 * This method finds prob(pos class) for each example. The bound and
	 * scoreCPT parameters allow us to find the probability of examples after
	 * adding a new feature to the network.
	 * 
	 * @param scoreSet
	 *            the data for which to find the probabilities
	 * @param bound
	 *            the number of features to use in computing the score
	 * @param scoreCPT
	 *            the cpts to use to find probabilities
	 * @return prob(pos class) of each example in scoreSet
	 */
	protected double[] getProbs(byte[][] scoreSet, int bound,
			ProbDistribution[] scoreCPT, int ignoreV1, int ignoreV2) {
		double cat1tmp = 1;
		double cat0tmp = 1;
		double norm, cat1Prob;
		double[] postProb = new double[scoreSet.length];

		for (int ix = 0; ix < scoreSet.length; ix++) {
			// Add in class prior
			cat1tmp = cat1Prior;
			cat0tmp = cat0Prior;
			// loop through the attributes
			for (int jx = 0; jx < bound; jx++) {
				// check whether we should ignore jx
				if (jx != ignoreV1 && jx != ignoreV2) {
					cat1tmp += scoreCPT[jx].getLogProb(scoreSet[ix],
							outputPosition, 1);
					cat0tmp += scoreCPT[jx].getLogProb(scoreSet[ix],
							outputPosition, 0);
				}
			}// end of jx loop through individual feature vector
			// get normalization constant
			norm = Math.exp(cat1tmp) + Math.exp(cat0tmp);
			// Calculate Probability
			postProb[ix] = (Math.exp(cat1tmp) / norm);
		}
		return postProb;
	}// end getProbs(byte[][], int, ProbDistribution[]);

	/**
	 * Initials the statistics for the graph structure for speculative models
	 * (i.e., the model used to evaluate feature additions, removals and
	 * replacements.
	 * 
	 * @param bound
	 *            much of the graph to initialize
	 */
	protected void initEvalGraph(int bound) {
		for (int ix = 0; ix < bound; ix++) {
			evalOutDegree[ix] = 0;
			evalInDegree[ix] = 0;
			evalParents[ix] = Constants.NO_PARENT;
			for (int jx = (ix + 1); jx < bound; jx++) {
				evalGraph[ix][jx] = Constants.NO_EDGE;
				evalGraph[jx][ix] = Constants.NO_EDGE;
			}// end jx loop
		}// end ix loop
	}// end initEvalGraph(int)

	protected void doubleSizes() {
		super.doubleSizes();
		double[][][] newCondProbs = new double[initSize][][];
		for (int ix = 0; ix < numRules; ix++) {
			newCondProbs[ix] = condProbs[ix];
		}
		int[] newParents = new int[initSize];
		for (int ix = parents.length; ix < newParents.length; ix++) {
			newParents[ix] = Constants.NO_PARENT;
		}
		System.arraycopy(parents, 0, newParents, 0, parents.length);

		Edge[][] newAllEdges = new Edge[initSize][initSize];
		// init graph
		int[] newInDegree = new int[initSize];
		int[] newOutDegree = new int[initSize];
		System.arraycopy(inDegree, 0, newInDegree, 0, numRules);
		System.arraycopy(outDegree, 0, newOutDegree, 0, numRules);
		for (int ix = 0; ix < numRules; ix++) {
			System.arraycopy(allEdges[ix], 0, newAllEdges[ix], 0, numRules);
		}
		condProbs = newCondProbs;
		parents = newParents;
		allEdges = newAllEdges;
		inDegree = newInDegree;
		outDegree = newOutDegree;

		// Speculatively network structure and info
		evalCPTs = new CPTByte[initSize];
		evalParents = new int[initSize];
		for (int ix = 0; ix < evalParents.length; ix++) {
			evalParents[ix] = Constants.NO_PARENT;
		}
		evalGraph = new int[initSize][initSize];
		evalInDegree = new int[initSize];
		evalOutDegree = new int[initSize];
	}
}// end class IncTan