package sayu;
/**
   This class encapsulates information about the performance of a
   classifier. It stores the fpr, tpr, recall and precision. It is used for finding the convex hull of a set of points in ROC space. 
 */
public class ROCPoint implements Comparable{

    public static final int THETA = 0;
    public static final int FPR = 1;
    public static final int RECALL = 2;


    private double fpr; //the false positive rate of this classifier
    private double tpr; //the true positive rate (recall) of this classifier
    private double precision; //the precision of this classificier
    private double theta; //the polar angle formed by this point in ROC space and the origin
    private double threshold;

    //the compare function to use
    private int compare = 0;

    /**
       This class encapsulates information about the performance of a
       classifier. It stores the fpr, tpr, recall and precision.
       @param fpr the false positive rate of this classifier
       @param tpr the true positive rate (and recall) of this classifier
       @param prec the precision of this classificier
     */
    public ROCPoint(double fpr, double tpr, double prec, double threshold){
        this.fpr = fpr;
        this.tpr = tpr;
        this.precision = prec;
        this.threshold = threshold;
        theta = Math.atan2(tpr, fpr);
    }//end of constructor

    /**
       This class encapsulates information about the performance of a
       classifier. It stores the fpr, tpr, recall and precision.
       @param fpr the false positive rate of this classifier
       @param tpr the true positive rate (and recall) of this classifier
       @param prec the precision of this classificier
     */
    public ROCPoint(double fpr, double tpr, double prec){
        this(fpr, tpr, prec, 0);
    }    

    public ROCPoint(double fpr, double tpr){
        this.fpr = fpr;
        this.tpr = tpr;
        theta = Math.atan2(tpr, fpr);
    }//end of constructor ROCPoint(double fpr, double tpr)

    /**
       @return the false positive rate of this classifier
     */
    public double getFPR(){
        return fpr;
    }//end of method getFPR()

    /**
       @return the true positive rate of this classifier
     */
    public double getTPR(){
        return tpr;
    }//end of method getTPR()

    /**
       @return the polar angle formed by this point in ROC space and the origin
     */
    public double getTheta(){
        return theta;
    }//end of method getTheta()

    /**
       @return the recall of this classifier.
     */
    public double getRecall(){
        return tpr;
    }//end of method getRecall()

    /**
       @return the precision of this classifier
     */
    public double getPrecision(){
        return precision;
    }//end of method getPrecision()

    /**
       @return the threshold used to obtain this classifier
     */
    public double getThreshold(){
        return threshold;
    }

    /**
       @return the PRPoint formed by this classifier
     */
    public PRPoint getPRPoint(){
        return new PRPoint(tpr, precision);
    }//end of method getPRPoint()


    /**
       Performs a comparision based on TPR (or recall).
       @param o is of type ROCPoint
       @return -1 if this.getTPR() < o.getTPR(). 0 if o.getTPR() == this.getTPR(). 1 if this.getTPR() > o.getTPR()
     */
    public int tprComp(Object o){
        ROCPoint tmp = (ROCPoint)o;
        double tpr2 = tmp.getTPR();
        if (tpr < tpr2){
            return -1;
        }
        else if (tpr > tpr2){
            return 1;
        }
        else{
            return 0;
        }
    }//end of method tprComp(Object o);

    /**
       Peforms a comparision based on FPR.
       @param o is of type ROCPoint
       @return -1 if this.getFPR() < o.getFPR(). 0 if o.getFPR() == this.getFPR(). 1 if this.getFPR() > o.getFPR()

     */
    public int fprComp(Object o){

        ROCPoint tmp = (ROCPoint)o;
        double fpr2 = tmp.getFPR();
        if (fpr < fpr2){
            return -1;
        }
        else if (fpr > fpr2){
            return 1;
        }
        else{
            double tpr2 = tmp.getTPR();
            if (tpr < tpr2){
                return -1;
            }
            else if (tpr > tpr2){
                return 1;
            }
            else{
                return 0;
            }
        }
    }//end of method fprComp(Object o);

    /**
       Peforms a comparision based on the polar angle formed by this point in ROC space and the origin.
       @param o is of type ROCPoint
       @return -1 a)this.getTheta() < o.getTheta() b) o.getTheta() == this.getTheta() && this.getFPR() < o.getFPR() c)  o.getTheta() == this.getTheta() && this.getFPR() == o.getFPR() && this.getTPR() < o.getTPR().  0 if o.getTheta() == this.getTheta() && this.getFPR() == o.getFPR() && this.getTPR() == o.getTPR().  1 if a) this.getTheta() > o.getTheta() b) this.getTheta() == o.getTheta() &&  this.getFPR() > o.getFPR() c)  o.getTheta() == this.getTheta() && this.getFPR() == o.getFPR() && this.getTPR() > o.getTPR().
     */
    public int thetaComp(Object o){
        ROCPoint tmp = (ROCPoint)o;
        double theta2 = tmp.getTheta();
        if (theta < theta2){
            return -1;
        }
        else if (theta > theta2){
            return 1;
        }
        else{
            double fpr2 = tmp.getFPR();
            if (fpr < fpr2){
                return -1;
            }
            else if (fpr > fpr2){
                return 1;
            }
            else{
                double tpr2 = tmp.getTPR();
                if (tpr < tpr2){
                    return -1;
                }
                else if (tpr > tpr2){
                    return 1;
                }
                else{
                    return 0;
                }
            }
        }
    }//end of method thetaComp(Object o);

    /**
       Sets the comparator to use with these points. 
       @param c the function to use to compare two ROCPoints
     */
    public void setCompare(int c){
        compare =c ;
    }//end of method setCompare()




    /**
       Peforms a comparision based on current setting of compare.
       @param o is of type ROCPoint
       @return -1 if this < o. if this == o. 1 if this > o.
     */
    public int compareTo(Object o){
        if (compare == FPR){
            return fprComp(o);
        }
        else if (compare == RECALL){
            return tprComp(o);
        }
        else{
            return thetaComp(o);
        }
    }//end of method compareTo(Object o);


    /**
       @return the false positive rate followed by a space followed by the true positive rate.
     */
    public String toString(){
        return "" + fpr + " " + tpr + " " + precision + " " + theta;
    }//end of method toString()
}//end of class ROCPoint
