
package sayu;
import java.util.*;
public class ScoreFunctions{

    static double minRecall = 0.4;

    public static void setAUCPRMinRecall(double new_min) {
	minRecall = new_min;
    }

    static double maxRecall = 1.0;

    public static void setAUCPRMaxRecall(double new_max) {
	maxRecall = new_max;
    }

    /**
       This method computes the conditional log likelihood of a set of
       examples.

       @param probs contains the Prob(Pos Class)
       @param cats containst the example's true class

       @return the summed condition log likelihood of the class variable
     */
    public static double cll(double[] probs, int[] cats){
	double cll = 0;
        for(int ix = 0; ix < probs.length; ix++){
            if (cats[ix] == 0){
		cll = cll + Math.log(1 - probs[ix]);
	    }
	    else{
		cll = cll + Math.log(probs[ix]);
	    }
	}
	return cll;
    }//end cll(double[], int[]);

   /**
       This method provides a fast way to recalculate the accuracy of
       a score set.  It is useful for picking a threshold for
       classifying something as positive. it is also useful for
       generating ROC or Precision/Recall curves.

       @param probs the probabilities of being cat1
       @param cats the classification of the example
       @param thres the threshold for calling example cat1

       @return the confusion matrix generated by calling any example with probability >= thres as a positive.
     */
    public static int[] fastAccuracy(double[] probs, int[] cats, double thres){
        int[] roc = new int[4];
        for(int ix = 0; ix < roc.length; ix++){
            roc[ix] = 0;
        }
        for(int ix = 0; ix < probs.length; ix++){
            if (probs[ix] >= thres){
		int pos = (cats[ix] == 1) ? 0: 1;
		roc[pos]++;
            }//end if
            else{
		int pos = (cats[ix] == 1) ? 2: 3;
		roc[pos]++;
            }//end else
        }//end for ix
        return roc;
    }//end method fastAccuracy
    /**
       This method calculates the precision/recall and ROC points for
       each possible classifier according to the probabilities and
       classifications defined by probs and cats.  It then takes the convex hull, in ROC space of these points and returns the classifiers on this hull. 

       @param probs the probabilities of being cat1
       @param cats the classification of the example
       @return the set of ROCPoints corresponding to convex hull of all possible classifiers that could be constructed
     */
    public static ROCPoint[] accuracyScoreAllSplits(double[] probs, int[] cats, boolean buildHull){
       
        int posCnt = 0;
	int negCnt = 0;
	HashMap<ClassSort, Integer> pos = new HashMap<ClassSort, Integer>();
	HashMap<ClassSort, Integer> neg = new HashMap<ClassSort, Integer>();
	Integer tmpInt;
	ClassSort sortObj;
        for(int ix = 0; ix < probs.length; ix++){
            sortObj = new ClassSort(probs[ix], cats[ix]);
	    if (cats[ix] == 1){
		posCnt++;
		tmpInt = pos.get(sortObj);
		if (tmpInt == null){
		    pos.put(sortObj, 1);
		}
		else{
		    pos.put(sortObj, (tmpInt+1));
		}
	    }
	    else{
		negCnt++;
		tmpInt = neg.get(sortObj);
		if (tmpInt == null){
		    neg.put(sortObj, 1);
		}
		else{
		    neg.put(sortObj, (tmpInt+1));
		}
	    }
        }
	int index = 0;
	ClassSort[] sort = new ClassSort[neg.size() + pos.size()];
	Iterator<ClassSort> iter;
	for(iter = pos.keySet().iterator(); iter.hasNext(); ){
	    sortObj = iter.next();
	    sortObj.setCount(pos.get(sortObj));
	    sort[index] = sortObj;
	    index++;
	}
	for(iter = neg.keySet().iterator(); iter.hasNext(); ){
	    sortObj = iter.next();
	    sortObj.setCount(neg.get(sortObj));
	    sort[index] = sortObj;
	    index++;
	}
        return accuracyScoreAllSplits(sort, posCnt, negCnt, buildHull);
    }

    public static ROCPoint[] accuracyScoreAllSplits(ClassSort[] sort, int pos, int neg, boolean buildHull){
	Arrays.sort(sort);
        
        double prevThres = sort[sort.length-1].getProb();
        double currThres;
        int prevClass = sort[sort.length-1].getClassification();
        int posSeen = 0;
	int negSeen = 0;
	if (prevClass == 1){
	    posSeen = sort[sort.length-1].getCount();
	}
	else{
	    negSeen = sort[sort.length-1].getCount();
	}
        int currClass, fp, fn, tn;
        
    
        LinkedList<ROCPoint> rocPoints = new LinkedList<ROCPoint>();
        double recall, prec, fpr;
        for(int ix = sort.length-2; ix >=0 ; ix--){
            currClass = sort[ix].getClassification();
            currThres = sort[ix].getProb();
            //class switch between this point and prev point
            if (prevClass == 1 && 0 == currClass){
                //lets think about this...
                fp = negSeen;
                fn = pos - posSeen;
                tn = neg - fp;		
                recall = ((double)posSeen)/((double)(posSeen + fn));
                prec = ((double)posSeen)/((double)(posSeen + fp));
                fpr = (double)fp/((double)(fp + tn));
                rocPoints.add(new ROCPoint(fpr, recall, prec, prevThres));

            }
	    if (currClass == 1){
		posSeen += sort[ix].getCount();
	    }
	    else{
		negSeen += sort[ix].getCount();
	    }

            prevThres = currThres;
            prevClass = currClass;
        }
        rocPoints.add(new ROCPoint(0,0,1));
        rocPoints.add(new ROCPoint(1,1,((double)pos)/((double)(pos+neg))));


        if (buildHull)
            return convexHull(rocPoints, ROCPoint.FPR);
        else{
            return toArrayROC(rocPoints);
        }
    }//end of method accuracyScoreAllSplits

    /**
       @param roc a linked list of ROCPoints
       @return an array representation of the ROCPoints
     */
    public static ROCPoint[] toArrayROC(LinkedList<ROCPoint> roc){
	ROCPoint[] pts = new ROCPoint[roc.size()];
	for(int ix = 0; ix < pts.length; ix++){
	    pts[ix] = roc.removeFirst();
	}
	return pts;
    }//toArrayROC(LinkedList<ROCPoint>)

    /**
       This method uses a graham scan to find the points that form a
       convex hull in ROC space.
       @param roc list of ROCPoint corresponding to different classifiers.
       @param pointType whether the returned points should be sorted
       as ROC or PR points.
       @return the points that form a convex hull in ROC space.
     */
    public static ROCPoint[] convexHull(LinkedList<ROCPoint> roc, int pointType){
	//convert to an array
	ROCPoint[] pts = toArrayROC(roc);
	//sort according to polar angle 
	Arrays.sort(pts);
	
	int prev = 0;
	double theta = pts[0].getTheta();
	boolean[] notdup = new boolean[pts.length];
	notdup[0] = true;
	int count = 1;
	double rand = Math.atan2(1,1);
	/**
	   remove duplicate points.  keep the duplicate that is furtherest from the origin.
	*/
	for(int ix = 1; ix < pts.length; ix++){
	    if (pts[ix].getTheta() < rand){
		//remove points that do worse than random guessing.
		notdup[ix] = false;
	    }
	    else{
		if (theta < pts[ix].getTheta()){
		    theta = pts[ix].getTheta();
		    count++;
		}
		else{
		    notdup[prev] = false;
		}
		prev = ix;
		notdup[ix] = true;
	    }
	}
	
	ROCPoint[] newpts = new ROCPoint[count];
	int pos = 0;
	for(int ix = 0; ix < pts.length; ix++){
	    if (notdup[ix]){ 
		newpts[pos] = pts[ix];
		pos++;
	    }	    
	}
	LinkedList<ROCPoint> stack = new LinkedList<ROCPoint>();
	if (newpts.length < 3){
	    for(int ix = 0; ix < newpts.length; ix++){
		stack.addFirst(newpts[ix]);
	    }
	}
	else{
	    for(int ix = 0; ix < 3; ix++){
		stack.addFirst(newpts[ix]);
	    }
	    for(int ix = 3; ix < newpts.length; ix++){
		while((stack.size() > 1 ) && nonLeftTurn(stack.get(1) , stack.get(0) , newpts[ix])){
		    ROCPoint tmpPT = stack.removeFirst();
		}
		stack.addFirst(newpts[ix]);
	    }
	}
	//Convert points in convex hull back into an array
	Object[] obj = stack.toArray();
	ROCPoint[] hull = new ROCPoint[obj.length];
	for(int ix = 0; ix < obj.length; ix++){
	    hull[ix] = (ROCPoint)obj[ix];
	    hull[ix].setCompare(pointType);
	}
	
	return hull;
    }


    /**

       @param nextTop the ROCPoint on the next from the top of the stack
       @param top the ROCPoint on the top of the stack
       @param pi the ROCPoint currently being evaluated
       @return true iff the angle formed by nextTop, top and pi make a non-left turn in ROC space
     */
    private static boolean nonLeftTurn(ROCPoint nextTop, ROCPoint top, ROCPoint pi){
        double cross = ((pi.getFPR() - top.getFPR()) * (top.getTPR() - nextTop.getTPR())) - (( top.getFPR() - nextTop.getFPR() ) * ( pi.getTPR() - top.getTPR() ));
        boolean ans = (cross < 0) ? false: true;
	return ans;
	
    }//end of method nonLeftTurn

   /**
       This calculates the area under an Precision Recall curve
       defined by the points contained in hull.  Note that the curve
       defined by hull is not guaranteed to be optimal p/r curve. This
       method was original written by Mark Goadrich. I have modified
       it to improve its efficiency.
       @param hull an set of ROCPoints (which have prec/recall info)
       @param display if true, write each point and AUC to standard out
       @param posTotal total number of positives in this dataset
       @param negTotal total number of negatives in this dataset
       @return results[0] = total area under the PR curve, results[1] = area under PR curve >= minRecall
     */
    public static double[] calculate(ROCPoint[] hull, int posTotal, int negTotal, double minRecallArea, boolean display){

	LinkedList<MyPoint> init = new LinkedList<MyPoint>();
        for(int ix = 0; ix < hull.length; ix++){
            double recall = hull[ix].getRecall();
            double precision = hull[ix].getPrecision();
            int positive = (int)(recall * posTotal);
            int negative = (int)((positive - (precision * positive))/ precision);
	    if (positive > 0){
		init.add(new MyPoint(positive, negative));
	    }
	}
	init.add(new MyPoint(posTotal, negTotal)); 
	MyPoint[] points = new MyPoint[init.size()];
	for(int ix = 0; ix < points.length; ix++){
	    points[ix] = init.removeFirst();
	}
        Arrays.sort(points);
        // calculate area under rp curve
	MyPoint curr = points[0];
        double rec = (curr.getPos() / (double)posTotal);
        double prec = (curr.getPos() / (curr.getPos() + (double)curr.getNeg()));

        if (display){
            System.out.println(0 + " " + prec);
            System.out.println(rec + " " + prec);
        }
        double area = rec * prec;
        boolean save = true, save2 = true;
        double tmpArea = 0, tmpArea2 = 0;
	double nrec, nprec, add1, add2;
	int pdiff, ndiff, currPos, currNeg, oldPos, oldNeg;
        // for each pair of points
        for (int i = 0; i < points.length - 1 ; i++) {
            // add intermediate anchors
            curr = points[i];
            MyPoint next = points[i+1];
            pdiff = next.getPos() - curr.getPos();
            currPos = curr.getPos();
            currNeg = curr.getNeg();
            ndiff = next.getNeg() - curr.getNeg();
            double margin = ndiff / (double)pdiff;
	    oldPos = currPos;
	    oldNeg = currNeg;
	    do{
                oldNeg = (oldPos == next.getPos() ||  oldPos == (next.getPos() - 1)) ? next.getNeg() :(int)(Math.round(currNeg + ((oldPos - currPos + 1 )* margin)));
		oldPos = (oldPos == next.getPos()) ? next.getPos() : oldPos + 1;
		nrec = (oldPos / (double)posTotal);
		nprec = (oldPos / (double)(oldPos + oldNeg)); 
		add1 =  (nrec - rec) * nprec;
		add2 =  0.5 * (nrec - rec) * (prec - nprec);
		if (save && nrec >= minRecallArea){
		    double sv = (rec >= minRecallArea) ? prec: solve(nrec, nprec, rec, prec, minRecallArea) ;
		    double a1 = (minRecallArea - rec) * sv;
		    double a2 = .5 * (minRecallArea - rec) * (prec - sv);
		    tmpArea = area + a1 + a2;                
		    save = false;
		}
		double maxRecallArea = maxRecall;
		if (save2 && nrec >= maxRecallArea){
		    double sv = (rec >= maxRecallArea) ? prec: solve(nrec, nprec, rec, prec, maxRecallArea) ;
		    double a1 = (maxRecallArea - rec) * sv;
		    double a2 = .5 * (maxRecallArea - rec) * (prec - sv);
		    tmpArea2 = area + a1 + a2;                
		    save2 = false;
		}
		area += add1 + add2;
		rec = nrec;
		prec = nprec;
		if (display){
		    System.out.println(rec + " " + prec);
		}
            }while (oldPos != next.getPos() && oldPos != next.getPos() + 1);

        }
        double[] results =new double[2];
        results[0] = area;
        results[1] = tmpArea2-tmpArea;
	//	System.err.println(results[0] + "/" + results[1]+"/" + (area-tmpArea));

        return results;

    }
   /**
       Finds the line determined by these two points and returns the
       y-value for for x input = val. This method is used to ensure
       that we can start integration in PR space at any level of
       recall.
       @param x1 x-coordinate of point 1
       @param y1 y-coordinate of point 1
       @param x2 x-coordinate of point 2
       @param y2 y-coordinate of point 2
       @param val - value for which to calculate the y-value
       @return slope(val) + b
     */
    public static double solveX(double x1, double y1, double x2, double y2, 
            double yTarget){
        double slope = (y1-y2)/(x1-x2);
        double b = y1 - (slope * x1);
        return (yTarget - b)/slope; 
    }//end solve

    /**
       Finds the line determined by these two points and returns the
       y-value for for x input = val. This method is used to ensure
       that we can start integration in PR space at any level of
       recall.
       @param x1 x-coordinate of point 1
       @param y1 y-coordinate of point 1
       @param x2 x-coordinate of point 2
       @param y2 y-coordinate of point 2
       @param val - value for which to calculate the y-value
       @return slope(val) + b
     */
    public static double solve(double x1, double y1, double x2, double y2, 
			       double val){
        double slope = (y1-y2)/(x1-x2);
        double b = y1 - (slope * x1);
        return (y1 - (slope * (x1 - val)));
    }//end solve

    /**
       Generates all possible PR points defined by the examples. Finds
       the optimal set of points in PR space and calculates the area
       under the PR curve.

       @param probs the probabilities of being cat1 
       @param cats the classification of the example
       @param posTotal total number of positives in this dataset
       @param negTotal total number of negatives in this dataset
       @return results[0] = total area under the PR curve, results[1] = area under PR curve >= minRecall

     */
    public static double[] calcPRArea(double[] probs, int[] cats, int posTotal, int negTotal){
        return calcPRArea(probs, cats, false, posTotal, negTotal);
    }

    /**
       Generates all possible PR points defined by the examples. Finds
       the optimal set of points in PR space and calculates the area
       under the PR curve.

       @param probs the probabilities of being cat1
       @param cats the classification of the example
       @param display if true, write each point and AUC to standard out
       @param posTotal total number of positives in this dataset
       @param negTotal total number of negatives in this dataset
       @return results[0] = total area under the PR curve, results[1] = area under PR curve >= minRecall
     */
    public static double[] calcPRArea(double[] probs, int[] cats, boolean display, int posTotal, int negTotal){
        ROCPoint[] hull = accuracyScoreAllSplits(probs, cats, false);
        return calculate(hull, posTotal, negTotal, minRecall, display);
    }

    public static double calcROCArea(double[] probs, int[] cats){
	ROCPoint[] hull = accuracyScoreAllSplits(probs, cats, false);
	return calcROCArea(hull, false, true);
    }

    public static double calcROCArea(double[] probs, int[] cats, boolean display){
	ROCPoint[] hull = accuracyScoreAllSplits(probs, cats, false);
	return calcROCArea(hull, display, true);
    }

    /**
       This calculates the area under an ROC curve defined by the
       points contained in hull.  Note that hull the ROC curve is not
       guaranteed to be a convex hull.
       @param hull an set of ROCPoints
       @param display if true, write each point and AUC to standard out
       @param setCompare if true, ensures that comparison between points is set to be false positive rate. (Note: default is set compare based on polar angle, for computing the convex hull)
       @return the area under the ROC curve defined by the points in hull
     */
    public static double calcROCArea(ROCPoint[] hull, boolean display, boolean setCompare){

        if (setCompare){
            for(int ix =0; ix < hull.length; ix++){
                hull[ix].setCompare(ROCPoint.FPR);
            }
        }
        Arrays.sort(hull);
        if (display){
            for(int ix = 0; ix < hull.length; ix++){
                System.out.println(hull[ix].toString());
            }
        }
        double prevFPR  = hull[0].getFPR();
        double prevTPR  = hull[0].getTPR();
        double currFPR = 0, currTPR = 0;
        double add1 = 0, add2 =0;
        double area = 0;
        for(int ix =1; ix < hull.length; ix++){
            currFPR = hull[ix].getFPR();
            currTPR = hull[ix].getTPR();
            double base = (currFPR - prevFPR);
            add1 = prevTPR * base;
            add2 = .5 * (base * ( currTPR - prevTPR ) );
            area = area + add1 + add2;
            prevTPR = currTPR;
            prevFPR = currFPR;
        }
        if (display){
            System.out.println(area);
        }
        return area;

    }//end of method

}
