%
% This module implements the linkage and aggregation operations. 
%  Linkage matches variables in the example with variables in the
%     newly invented predicate,
%  Aggregation allows replacing a clause body by some aggregate on a variable. 
% 
% Interface:
%  aggregation/3: InventedPredicate, LinkedToExampleClause, 
% WhetherToUseAggregates.
%
% Global Variables:
%
%  aggregation_type: what types of agg are supported, values: aggs, single
%

:- module(vista_aggregation, [aggregation/3,
	aggregated/4,
	linkable/1,
	linkage/2,
	linkage_compatible/1,
	linkage/3]).

:- use_module(library(terms),[variable_in_term/2]).

:- use_module(library(lists),[member/2,
	subtract/3]).

:- use_module(modes,[determination/1,
		     determination_arity/1,
		     modeh/1,
		     modeb/1]).

:- use_module(examples,[new_predicate_to_call/1]).

:- dynamic aggregation_type/1.

aggregation_type(aggs).

aggregation((Head :- Body), (Example :- LinkedAggBody), UseAggregates) :- !,
	% do aggregation
	aggregated(Body, Head, AggBody, UseAggregates),
	% do linkage
	link(Head, Example, AggBody, LinkedAggBody).
aggregation(_Head, Example, _) :-
	determination(Example).

%
% linkage data-base
%
:- dynamic linkage/2, linkage/5, link_compat/1.

%
% ensure there is a way to link this with the clause_body.
%
linkable(Head) :-
	determination(Example),
	!,
	link(Head, Example, true, _).

link(Head, Example, AggBody, LinkedAggBody) :-
	functor(Head,PName,PArity),
	functor(PredMode,PName,PArity),
	modeb(PredMode),
	modeh(Modeh),
	ensure_linkage_compat(PredMode),
	functor(Modeh, EName, EArity),
	functor(Example, EName, EArity),
	link_args(0, PArity, Head, PredMode, Example, Modeh, [], AggBody, LinkedAggBody).

link_args(PArity, PArity, _, _, _, _, _, B, B) :- !.
link_args(I0, PArity, Head, PredMode, Example, Modeh, ArgsUsed, B0, BF) :-
	I is I0+1,
	arg(I, PredMode, ArgMode),
	arg(I, Head, Arg),
	link_arg(ArgMode, Arg, I, Example, Modeh, ArgsUsed, Link, NewArgsUsed, ArgsUsed),	
	add_linkage(Link, B0, BI),
	link_args(I, PArity, Head, PredMode, Example, Modeh, NewArgsUsed, BI, BF).

link_arg(-_, _, _, _, _, _, true) --> [].
link_arg(+Type, Arg, I, Example, Modeh, ArgsUsed, Link) -->
	[I],
	link_this_arg(Type, Arg, I, Example, Modeh, ArgsUsed, Link), !.
link_arg(+Type, Arg, _, Example, Modeh, ArgsUsed, Link) -->
	{ functor(Example, _, Ar), between(1, Ar, J) },
	[J],
	link_this_arg(Type, Arg, J, Example, Modeh, ArgsUsed, Link), !.


link_this_arg(Type, Arg, I, Example, ModeH, ArgsUsed, Link) -->
	% do not allow an example argument to be used twice
	{ available(I, ArgsUsed),
	% get the mode of the head
	  arg(I, ModeH, MArg),
	% get the type from the mode
	  arg(1, MArg, TArg),
	  link(TArg, Type, Link, ExArg, Arg),
	  arg(I, Example, ExArg)
        }.

available(_, []).
available(I, J.L) :-
	J \= I,
	available(I, L).

add_linkage(true, AggBody, AggBody) :- !.
add_linkage((A,B), AggBody, LinkedAggBody) :- !,
	add_linkage(A, AggBody, LinkedAggBody1),
	add_linkage(B, LinkedAggBody1, LinkedAggBody).
add_linkage(Linkage, AggBody, (Linkage,AggBody)).

linkage_compatible(Modes0) :-
	sort(Modes0,Modes),
	assert(link_compat(Modes)).

ensure_linkage_compat(Modeh) :-
	Modeh =.. [_|Modes],
	strip(Modes,Mode1.MyModes),
	link_compat(IModes),
	% allow a type to appear in other modes
	member(Mode1, IModes),
	!,
	once((
	      link_compat(LinkModes),
	      subtract(Mode1.MyModes, LinkModes, []))).
ensure_linkage_compat(_).

strip([],[]).
strip((+M).Modes,M.MyModes) :-
	strip(Modes,MyModes).
strip((-M).Modes,M.MyModes) :-
	strip(Modes,MyModes).
strip((#M).Modes,M.MyModes) :-
	strip(Modes,MyModes).


    
link(Type,Mode,Ex,Link,Arg) :-
	linkage(Type,Mode,Ex,Link,Arg).

% fix this so if only exist then cut, else no cut
% existencial aggregation (the way Prolog computes things).
% other aggregations
aggregated(Body, Head, Agg, aggs) :-
	functor(Head,_,2),
	determination_arity(1),
	arg(1,Head,A1),
	agg_ops(AggType),
	% pruning and aggregates do not match
	remove_pruning(Body,B),
	% now we need to find out which variables we're aggregating on
	% these must be the variables in B, not in the example
	new_variables(B, [A1], Vs),
	Agg = aggregate(_,AggType,Vs,B).
aggregated(Body, Head, Agg, aggs) :-
	%write(step1),nl,
	functor(Head,_,3),
	determination_arity(2),
	arg(1, Head, A1),
	arg(2, Head, A2),
	agg_ops(AggType),
	% pruning and aggregates do not match
	remove_pruning(Body,B),
	new_variables(B, [A1,A2], Vs),
	Agg = aggregate(_,AggType,Vs,B).
aggregated(Body, _, Body, single ) :-
	(aggregation_type(existence) ->
	    !
	;
	    true
	).

remove_pruning('!'(A,B),(A,B)) :- !.
remove_pruning(B,B).

agg_ops(count).
	
new_variables(B, L, Vs) :-
 	term_variables(B, Vs0),
	extract_head_vars(Vs0, L, Vs).

extract_head_vars([], _, []).
extract_head_vars(V.Vs0, L, Vs) :-
	variable_in_term(L, V), !,
	extract_head_vars(Vs0, L, Vs).
extract_head_vars(V.Vs0, L, V.Vs) :-
	extract_head_vars(Vs0, L, Vs).
	
linkage(Type1, Type2, How) :-
	S =.. [How, X1, X2],
	new_predicate_to_call(S),
	retractall(linkage(Type1, Type2, _, _, _)),
	retractall(linkage(Type1, Type2)),
	assert(linkage(Type1, Type2, S, X1, X2)),
	assert(linkage(Type1, Type2)).
