%
% This module implements a simple clause optimiser. It assumes H is ground
%     at call and separates the clause in independent components, while removing
%     redundant calls.
%
%%
% Interface:
%  optimise_clause/2
%
% Global Variables:
%
% none
%
%

:- module(vista_optimise_clause,[optimise_clause/2]).

:- use_module(library(ordsets),[ord_subtract/3,ord_disjoint/2,ord_union/3,ord_subset/2]).
				%
:- use_module(library(lists),[member/2]).
				%
:- use_module('../common/utils',[conjunction_to_list/2,list_to_conjunction/2]).
				%
% Transform a clause for fast evaluation
%

optimise_clause(Clause,Clause1):-
	fail,
	remove_redundant(Clause,Clause0),
	reorder_clause(Clause0,Clause1,_Clause2), !.
optimise_clause(Clause,Clause).

remove_redundant((Head:-Body),(Head1:-Body1)):-
	conjunction_to_list((Head,Body),ClauseL),
	remove_subsumed(ClauseL,[Head1|Body1L]),
	(Body1L = [] -> Body1 = true; list_to_conjunction(Body1L,Body1)).

reorder_clause((Head:-Body), Clause, EndClause) :-
	term_variables(Head,LHead),
%	vars_in_term([Head],[],LHead),
	number_goals_and_get_vars(Body,LHead,1,IF,[],Conj),
	calculate_independent_sets(Conj,[],BSets),
	compile_clause(BSets,Head,Clause,IF,EndClause).

% make sure we cope with
number_goals_and_get_vars((G,Body),LHead,I0,IF,L0,LGs) :- !,
	number_goals_and_get_vars(G,LHead,I0,I1,L0,LI),
	number_goals_and_get_vars(Body,LHead,I1,IF,LI,LGs).
number_goals_and_get_vars(G,LHead,I0,I,L0,[g(I,LVF,NG)|L0]) :-
	I is I0+1,
	get_goal_vars(G,LHead,LVF,NG).

get_goal_vars(G,LHead,LVF,G) :-
	term_variables(G,LVI),
%	vars_in_term([G],[],LVI),
	ord_subtract(LVI,LHead,LVF).

calculate_independent_sets([],BSets,BSets).
calculate_independent_sets([G|Ls],BSets0,BSetsF) :-
	add_goal_to_set(G,BSets0,BSetsI),
	calculate_independent_sets(Ls,BSetsI,BSetsF).

add_goal_to_set(g(I,LV,G),Sets0,SetsF) :-
	add_to_sets(Sets0,LV,[g(I,LV,G)],SetsF).

add_to_sets([],LV,Gs,[[LV|Gs]]).
add_to_sets([[LV|Gs]|Sets0],LVC,GsC,[[LV|Gs]|SetsF]) :-
	ord_disjoint(LV,LVC), !,
	add_to_sets(Sets0,LVC,GsC,SetsF).
add_to_sets([[LV|Gs]|Sets0],LVC,GsC,SetsF) :-
	ord_union(LV,LVC,LVN),
	join_goals(Gs,GsC,GsN),
	add_to_sets(Sets0,LVN,GsN,SetsF).

join_goals([],L,L):- !.
join_goals(L,[],L):- !.
join_goals([g(I1,VL1,G1)|T],[g(I2,VL2,G2)|T2],Z) :-
	I1 < I2, !,
	Z = [g(I1,VL1,G1)|TN],
	join_goals(T,[g(I2,VL2,G2)|T2],TN).
join_goals([H|T],[g(I2,VL2,G2)|T2],Z) :-
	Z = [g(I2,VL2,G2)|TN],
	join_goals(T,[H|T2],TN).

compile_clause(Goals, Head, (Head:-Body), IF, (Head:-EndBody)):-
	compile_clause2(Goals, Body, IF, EndBody).

compile_clause2([[_|B]], B1, I, BL):-
	!,
	glist_to_goals(B, B1, I, B1, BL).
compile_clause2([[_|B]|Bs], '!'(B1,NB), I, BF):-
	glist_to_goals(B, B1, I, B1, BF),
	compile_clause2(Bs, NB, I, BF).

glist_to_goals([g(I,_,Goal)],Goal, I1, GT, GL) :- !,
	last_sub(I, GL, I1, GT).
glist_to_goals([g(I,_,Goal)|Goals],(Goal,Goals1), I1, GT, GL):-
	last_sub(I, GL, I1, GT),
	glist_to_goals(Goals,Goals1, I1, GT, GL).

last_sub(I,G1,I,GF) :- !, G1 = GF.
last_sub(_,_,_,_).


% remove literals subsumed in the body of a clause
remove_subsumed([Head|Lits],Lits1):-
	reverse_delete(Lit,Lits,Left), % modifed FPD
	\+(\+(redundant(Lit,[Head|Lits],[Head|Left]))), !,
	remove_subsumed([Head|Left],Lits1).
remove_subsumed(L,L).

reverse_delete(L,[A|B],A.C) :-
	reverse_delete(L,B,C).
reverse_delete(L,L.B,B).


% determine if Lit is subsumed by a body literal
redundant(Lit,Lits,[Head|Body]):-
	copy_term([Head|Body],Rest1),
	member(Lit1,Body),  % modified FPD
	Lit = Lit1,
	aleph_subsumes(Lits,Rest1).

aleph_subsumes(Lits,Lits1):-
	\+(\+((numbervars(Lits,0,_),numbervars(Lits1,0,_),ord_subset(Lits,Lits1)))).

