%
% mode to fetch a new literal from the bottom-clause.
%
% It is made more complicated because we have to emulate the foil algorithm
% and be compatible with its ground representation.
%
:- module(vista_mdie,[generate_mdie_clause/2]).

:- use_module('bottom_clause',[next_literal/3,
	ensure_sat/2,
	bottom_clause_to_clause/1,
	get_bc_literal/2]).

:- use_module('modes',[modeb/1,
	determining/2]).

:- use_module('../common/utils',[conjunction_to_list/2,
	run_goals/1]).

:- use_module(library(lists),[append/3]).

:- use_module(aggregation,[aggregation/3]).

:- use_module(reduce,[current_seed/3]).

:- use_module(variabilise,[variabilise/2,
	unvariabilise/2]).

:- use_module(library(tries),[trie_check_entry/3,
	trie_put_entry/3,
	trie_open/1,
	trie_close/1]).

generate_mdie_clause(GCl, GNewClause) :-
	trie_open(Trie),
	(
	    generate_mdie_clause_with_trie(GCl, Trie, GNewClause)
	;
	    trie_close(Trie), fail
	).

generate_mdie_clause_with_trie(GCl, Trie, GNewClause) :-
	variabilise(GCl, Cl),
	generate_mdie_clause_vars(Cl, NewClause),
	\+ trie_check_entry(Trie,NewClause,_),
	trie_put_entry(Trie,NewClause,_),
	unvariabilise(NewClause, GNewClause).

%
% Clauses from the last refinement step
%
% assumes Head is all input mode.
%
generate_mdie_clause_vars((Head :- Literals), (NewHead :- NewLiterals)) :- !,
	conjunction_to_list(Literals, L2),
	length(L2, ClLength),
	vista_search_space:clauselength(Max), Max > ClLength+1,
	aggregation((Head :- Literals), (NHead :- LinkedLits), single),
	copy_term(Head-NHead-Literals, NewHead-NewNHead-OldLiterals),
	conjunction_to_list(LinkedLits, L1),
	append(Link,L2,L1),
	current_seed(Id,PosNeg,NHead),
	ensure_sat(Id,PosNeg),
	run_goals(Link),
	match_bottom_clause(L2, 0, Ids),
	next_literal(Ids, _, (NewNHead :- NewLiterals)),
	match_lits(NewLiterals, OldLiterals).
generate_mdie_clause_vars((OHead), (OHead :- Lit)) :-
	% work on a copy
	copy_term(OHead, Head),
	% dirty trick to force expansion.
	aggregation((Head :- ugh), (NHead :- LinkedLits), single),
	conjunction_to_list(LinkedLits, L1),
	append(Link,[_],L1),
	current_seed(Id,PosNeg,NHead),
	ensure_sat(Id,PosNeg),
% (bottom_clause_to_clause(Cl) -> portray_clause(Cl) ; true),
	run_goals(Link),
	match_against_modes(Head, OHead, Lit).

match_bottom_clause([], _, []).
match_bottom_clause(G.Gs, I0, (Id-ModeL).Ids) :-
	match_goal_to_bottom_clause(G, Id),
	functor(G,Na,Ar),
	functor(Modeb,Na,Ar),
	modeb(Modeb),
	Modeb =.. [_|ModeL],
	Id > I0, 
	match_bottom_clause(Gs, Id, Ids).
	
match_goal_to_bottom_clause(G, Id) :-
	get_bc_literal(Id,G).

match_lits((L,NLiterals), (L,OldLiterals)) :- !,
	match_lits(NLiterals, OldLiterals).
match_lits((L,_), L).


match_against_modes(Head, OHead, Lit) :-
	modeb(Modeb),
	OHead =.. [_|Args],
	Head =.. [_|Vals],
	Modeb =.. [_|Modes],
	get_literal(Modes, Args, Vals, Lit).

get_literal(HModes, Args, Vals, Lit) :-
	determining(_,Modeb),
	modeb(Modeb),
	add_det(Modeb, HModes, Args, Vals, Lit).

add_det(Modeb, HModes, Args, Vals, Lit) :-
	Modeb =.. [Na|MArgs],
	satisfy_args(MArgs, HModes, Args, Vals, Lits, CLits, HasConstants),
	CLit =.. [Na|CLits],
	Lit =.. [Na|Lits],
	get_bc_literal(_,CLit),
	(HasConstants = yes -> true ; !).

satisfy_args([], _, _, _, [], [], no).
% if input needs a bound argument
satisfy_args(+T.MArgs, Ms, As, Vs, A.Args, V.Vals, HC) :-
	satisfy_arg(T, Ms, As, Vs, A, V),
	satisfy_args(MArgs, Ms, As, Vs, Args, Vals, HC).
% anything will do.
satisfy_args(-_.MArgs, Ms, As, Vs, _.Args, _.Vals, HC) :-
	satisfy_args(MArgs, Ms, As, Vs, Args, Vals, HC).
% we have a constant copied to the goal.
satisfy_args(#_.MArgs, Ms, As, Vs, A.Args, A.Vals, yes) :-
	satisfy_args(MArgs, Ms, As, Vs, Args, Vals, _).
	
satisfy_arg(T, +T._, A._, V._, A, V).
satisfy_arg(T, _.Ms, _.As, _.Vs, A, V) :-
	satisfy_arg(T, Ms, As, Vs, A, V).

copy_constants([], _, _, no).
copy_constants(#_.MArgs, V.CLit, V.Lit, yes) :-
	copy_constants(MArgs, CLit, Lit, _).
copy_constants(+_.MArgs, _.CLit, _.Lit, Stat) :-
	copy_constants(MArgs, CLit, Lit, Stat).
copy_constants(-_.MArgs, _.CLit, _.Lit, Stat) :-
	copy_constants(MArgs, CLit, Lit, Stat).
