%
% This module provides a generic interface to search space operations. 
% 
% It supports 4 types of search space:
%  heap, beam, bestfirst (like heap but using assert), and fifo.
%
% Interface:
%  init_search_space/0: virtual class to initialise search space representation
%
%  add_node_to_search_space/2: virtual class to store a new node with numeric 
% score N and clause Cl.
%
%  get_node_from_search_space/1: virtual class to retrieve a node with Clause
% from the search space.
%
%
% Global Variables:
% search: takes values beam, heap, bestfirst, fifo.
% minScore: unused.
%
%
:- module(vista_search_space,[init_search_space/0,
			      add_node_to_search_space/2,
			      get_node_from_search_space/1]).

:- use_module(beam,[init_beam/0,
			      add_to_beam/2,
			      get_from_beam/1]).

:- use_module(heap,[init_heap/0,
		    add_to_heap/2,
		    get_from_heap/1]).

:- use_module(bestfirst, [init_bestfirst/0,
			  update_beam/2,
			  get_clause_beam/1]).

:- use_module('../scorer/scorer', [current_score/1]).

:- use_module('../common/utils', [clause_length/2]).

:- dynamic search/1.

%
% can be one of beam, heap, fifo, bestfirst.
%
search(heap).

:-dynamic(minScore/1).

minScore(0.0).

:-dynamic(clauselength/1).

clauselength(4).

init_search_space :-
	search(Type),
	init_search(Type).

init_search(beam) :-
	init_beam.
init_search(heap) :-
	init_heap.
init_search(bestfirst) :-
	init_bestfirst.
init_search(fifo).

add_node_to_search_space(Scores, Clause) :-
	clause_length(Clause, Length),
	clauselength(MaxLength),
	Length =< MaxLength,
	numeric_score(Scores, Length, Score),
	search(Type),
	add_node_to_search_space(Type, Score, Clause).

numeric_score(Score, _, ScoreF) :- 
	number(Score), !, 
	ScoreF is -Score.
numeric_score(Scores, ClLength, ScoreF) :-
	Scores = [_|_],
	current_score(Scores0),
	totdiff(Scores, Scores0, 0.0, Diff),
%	Diff1 is Diff-(Diff*(1-random))/2,
	Diff1 is Diff-ClLength,
	ScoreF is -Diff1.

totdiff([], [], D, D).
totdiff(S.Scores, S0.Score, D0, Diff) :-
	D is (S-S0)/S0,
	D1 is D0+D,
	totdiff(Scores, Score, D1, Diff).


add_node_to_search_space(beam, Score, Clause) :-
	add_to_beam(Score,Clause).
add_node_to_search_space(heap, Score, Clause) :-
	clause_length(Clause, Length),
%	X is random,
%	add_to_heap(Length-Score,Clause).
	NScore is -Score,
	add_to_heap(Length-NScore,Clause).
add_node_to_search_space(bestfirst, Score, Clause) :-
	update_beam(Score,Clause).
add_node_to_search_space(fifo, _, Clause) :-
	recordz(to_refine,Clause,_).

get_node_from_search_space(Clause) :-
	search(Type),
	get_node_from_search_space(Type,Clause).

get_node_from_search_space(beam,Clause) :-
	get_from_beam(Clause).
get_node_from_search_space(heap,Clause) :-
	get_from_heap(Clause).
get_node_from_search_space(bestfirst,Clause) :-
	get_clause_beam(Clause).
get_node_from_search_space(fifo,Clause) :-
	recorded(to_refine,Clause,DBRef),
	erase(DBRef).
