%
% This utility module encapsulates access to sets of clauses, or theories. 
% Clauses have two representations:
%  a variable based, like what you expect in Prolog
%  a ground  representation, where each argument is what you expect.
%
% The module has standard interface: add more clauses, obtain
% a clause, dump a theory, replace a clause.
% 
% Interface:
%  add_to_theory/2: add a clause to the theory
%
%  extract_theory/2: unify first argument with current theory in
%     variabilised form, and second argument with ground version.
%
%  list_theory/0: dump current theory in standard error.
%
%  provar_clause/1: access ground representation.
%
%  theory_varclause/2: access clause Id in theory (variabilised form).
%
%  replace_clause_in_theory/3: replace clause Id by new clause Cl/GroundCl
%
%  add_node_to_search_space/2: virtual class to store a new node with numeric 
%      score N and clause Cl.
%
%  get_node_from_search_space/1: virtual class to retrieve a node with Clause
%     from the search space.
%
%
% Global Variables:
%  none.
%
%

:- module(vista_theory,[add_to_theory/2,
			list_theory/0,
			extract_theory/2,
			provar_clause/1,
			theory_varclause/2,
			replace_clause_in_theory/3,
			theory_size/1]).

:- use_module('../common/debug_write',[debug_format/3,
				       verbosity/1]).

theory_size(N) :-
	predicate_property(idb:theory_clause,number_of_clauses(N)), !.
theory_size(0).

add_to_theory(Clause, ProvarClause) :-
	recordz(theory_clause, Clause, _),
	recordz(provar_clause, ProvarClause, _).

% extract a theory from SAYU or VISTA
extract_theory(Theory, ProvarTheory) :-
	findall(Clause, recorded(theory_clause,Clause,_), Theory),
	findall(Clause2, recorded(provar_clause,Clause2,_), ProvarTheory).

provar_clause(Clause) :-
	recorded(provar_clause,Clause,_).

list_theory :-
	debug_format(1, '% ====================~n',[]),
	fail.
list_theory :-
	verbosity(V),
	V >= 1,
	recorded(theory_clause,Clause,_),
	Clause \= fail,
	portray_clause(user_error, Clause),
	Clause = (_ :- C,_),
	functor(C,A,1),
	atom_concat('hierarchy_',_,A),
	clause(user:C,_),
	portray_clause(user_error, C),
	fail.
list_theory :-
	verbosity(V),
	V >= 1,
	recorded(disjunct_list,D,_),
	DT =.. [D,_],
	clause(user:DT,true),
	portray_clause(user_error, DT),
	fail.
list_theory :-
	debug_format(1, '% ====================~n~n~n',[]).


theory_varclause(I,Clause) :-
	findall(Clause, recorded(theory_clause,Clause,_),Clauses),
	nth(I, Clauses, Clause).

% use backtracking to save space
replace_clause_in_theory(Id, Cl, _) :-
	% don't actually remove the clause, just mark it as removed.
	findall(Clause, (recorded(theory_clause,Clause,R),erase(R)), Theory),
	replace_nth(Id, Theory, Cl, NTheory),
	include_clauseset(NTheory, theory_clause),
	fail.
replace_clause_in_theory(Id, _, GCl) :-
	findall(Clause, (recorded(provar_clause,Clause,R),erase(R)), Theory2),
	replace_nth(Id, Theory2, GCl, NTheory2),
	include_clauseset(NTheory2, provar_clause),
	fail.
replace_clause_in_theory(_,_,_).

replace_nth( 1,[_|Th],Cl, [Cl|Th]) :- !.
replace_nth(N1,[T|Th],Cl,[T|NTh]) :-
	N is N1-1,
	replace_nth(N,Th,Cl,NTh).

include_clauseset([], _).
include_clauseset([C|Theory], Key) :-
	recordz(Key, C, _),
	include_clauseset(Theory, Key).

	
